/*
    The Language for Asynchronous Research and Development -- LARD
    Copyright (C) 1996-1999 AMULET Group, Department of Computer Science
        The University of Manchester, Oxford Road, Manchester, UK, M13 9PL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Id: ccheck.c,v 1.1.1.1 1999/07/15 18:47:31 bardslea Exp $


	Class checking and handling code

	11 Jun 1999, AMULET Group
*/

#include <alloc.h>
#include <error.h>
#include <simple.h>

#include "debug.h"
#include "parser_structs.h"
#include "scopes.h"
#include "dump.h"
#include "main.h"
#include "decls.h"
#include "ccheck.h"

bool check_arg_class ( struct expr * actual, struct expr * formal,
                       int argnum, bool stop_on_error );
struct expr * expand ( struct expr * expr );
void assoc_args ( struct list * pargs, struct list * aargs );
void assoc_arg ( struct expr * parg, struct expr * aarg );
bool args_classes_match ( struct list * actuals, struct list * formals );
struct list * mksubst_args ( struct list * args );



bool check_arg_classes ( struct list * actuals, struct list * formals,
                         int argnum, bool stop_on_error )
{
  if (actuals && formals)
    {
      return(check_arg_class(actuals->expr,formals->expr,argnum,stop_on_error)
             && check_arg_classes(actuals->next,
                                  formals->next,argnum+1,stop_on_error));
    }
  else if (actuals || formals)
    {
      if (stop_on_error)
      {
        fatal_error(0,"parameter count mismatch");
      }
      return(FALSE);
    }
  return(TRUE);
}


bool check_arg_class ( struct expr * actual, struct expr * formal, int argnum,
                       bool stop_on_error )
{

      /* Check that the class of the actual matches the class required
         by the formal. The class of the actual was found when it was
         bound.  The class required by the formal is found now from
         the declaration of the formal. */

      /* Formals record the coresponding actual so that "type
         parameters" can be checked. */

  formal->inst=actual;

      /* Check if the classes of the actual and formal match. */
  
  if (class_match(&(actual->class),get_class(formal)))
    {
      debug(debug_pass_2b,"Pass ");
    }

      /* If they don't match they may be forced to match by means of
         an implicit eval() or read() expression. */
  
  else if (class_xform(actual,formal))
  {
    debug(debug_pass_2b,"Pass ");
  }
  else
    {
      debug(debug_pass_2b,"Fail\n");
      if (stop_on_error)
      {
        nonfatal_error(actual->line,
                       "Parameter class mismatch (arg %d of %s); expecting ",
                       argnum, formal->parent->u.fc.ident->name);
        dump_expr(get_class(formal),-1);
        fprintf(stderr,", got ");
        dump_expr(actual->class,-1);
        fprintf(stderr,"\n");
		exit (1);
      }
	  return(FALSE);
    }
  return(TRUE);
}


bool class_match ( struct expr ** actualp, struct expr * formal )
{
  struct expr * actual;

  actual=*actualp;
  
  if (debug_pass_2d)
  {
    fprintf(stderr,"matching ");
    dump_expr(actual,-1);
    fprintf(stderr," (decl=%p) and ",actual->decl );
    dump_expr(formal,-1);
    fprintf(stderr," (decl=%p)\n",formal->decl);
  }
  
      /* Check for a formal that is an implicit parameter. */

  if (formal->sort==fcall_es && formal->u.fc.ident==decl2_ident)
  {
    formal->inst=actual;
    return(class_match(&(actual->class),get_class(formal)));
  }

      /* Check for record declarations, which never match */ /* OLD */
  /*  
  if ((formal->sort==fcall_es && formal->u.fc.ident==dot_ident)
      || (actual->sort==fcall_es && actual->u.fc.ident==dot_ident))
    internal_error("old style records encountered\n");
  */  

  if (!formal->decl)
    fatal_error(formal->line,"Umm... this formal has no declaration (?!)");
  
      /* Check for a formal that is a "type parameter". */

  if (formal->sort==fcall_es && formal->decl->inst)
    return(class_match(actualp,formal->decl->inst));

      /* Otherwise check for matching declarations and matching
         parameters. */
  
  if (actual->decl==formal->decl)
  {
    if (actual->sort==fcall_es)
      return(args_classes_match(actual->u.fc.args,formal->u.fc.args));
    else if (actual->sort==num_es)
      return (actual->u.n==formal->u.n);
    else if (actual->sort==str_es)
      return (actual->u.s==formal->u.s);
    else if (actual->sort==char_es)
      return (actual->u.s==formal->u.s);
    else
      internal_error("shouldn't be here");
  }
  
      /* Check for an actual that is a type alias */
  if (get_defn(actual->decl))
  {
    (*actualp)=expand(actual);
    (*actualp)->real=actual;
/*    fprintf(stderr,"Set real pointer for ");
    dump_expr(*actualp,-1);
    fprintf(stderr," to ");
    dump_expr(actual,-1);
    fprintf(stderr,"\n"); */
    return(class_match(actualp,formal));
  }
  return(FALSE);
}


struct expr * expand ( struct expr * expr )
{
  struct expr * decl;
  struct expr * proto;
  struct expr * defn;

  decl=expr->decl;
  proto=get_proto(decl);
  defn=get_defn(decl);
  assoc_args(proto->u.fc.args,expr->u.fc.args);
  return(mksubst(defn));
}


void assoc_args ( struct list * pargs, struct list * aargs )
{
  if (pargs && aargs)
  {
    assoc_arg(pargs->expr,aargs->expr);
    assoc_args(pargs->next,aargs->next);
  } else if (pargs || aargs)
    internal_error("confused by arg count");
}


void assoc_arg ( struct expr * parg, struct expr * aarg )
{
  parg->inst=aarg;
}


bool args_classes_match ( struct list * actuals, struct list * formals )
{
  if (actuals&&formals)
    return(class_match(&(actuals->expr),formals->expr)
	   && args_classes_match(actuals->next,formals->next));
  else if (actuals||formals)
    return(FALSE);
  else
    return(TRUE);
}


bool class_xform ( struct expr * actual, struct expr * formal )
{
      /* Check for var(X) when val(X) required, and expr(X) when
         val(X) required.  Set the class_xform field. */

/* hardcoded var, val and expr here is not good */

  if (!actual->class->decl)
    return(FALSE);
  
  if (get_proto(actual->class->decl)->u.fc.ident==putsym("var")
      && get_proto(get_class(formal)->decl)->u.fc.ident==putsym("val"))
  {
    if (class_match(&(actual->class->u.fc.args->expr),
                    arg(1,get_class(formal))))
    {
      actual->class_xform=cx_read;
      debug(debug_pass_2b,"(read) ");
      return(TRUE);
    }
  } else if (get_proto(actual->class->decl)->u.fc.ident==putsym("expr")
      && get_proto(get_class(formal)->decl)->u.fc.ident==putsym("val"))
  {
    if (class_match(&(actual->class->u.fc.args->expr),
                    arg(1,get_class(formal))))
    {
      actual->class_xform=cx_eval;
      debug(debug_pass_2b,"(eval) ");
      return(TRUE);
    }
  } else if (get_proto(actual->class->decl)->u.fc.ident==putsym("val")
      && get_proto(get_class(formal)->decl)->u.fc.ident==putsym("expr"))
  {
    if (class_match(&(actual->class->u.fc.args->expr),
                    arg(1,get_class(formal))))
    {
      actual->class_xform=cx_mkexpr;
      debug(debug_pass_2b,"(mkexpr) ");
      return(TRUE);
    }
  } else if (get_proto(actual->class->decl)->u.fc.ident==putsym("var")
      && get_proto(get_class(formal)->decl)->u.fc.ident==putsym("expr"))
  {
    if (class_match(&(actual->class->u.fc.args->expr),
                    arg(1,get_class(formal))))
    {
      actual->class_xform=cx_mkexpr_read;
      debug(debug_pass_2b,"(mkexpr read) ");
      return(TRUE);
    }
  }
  return(FALSE);
}


void get_return_class ( struct expr * expr )
{

      /* The return class for an expression is found from the class of
         its declaration, with any necessary substitutions from the
         actuals coresponding to the parameters. */

  expr->class=mksubst(get_class(expr->decl));
  if (debug_pass_2b)
    dump_expr(expr->class,-1);
}


struct expr * mksubst ( struct expr * expr )
{
  struct expr * p;
  
      /* Substitute an actual when the expression is a parameter. */

  if (expr->sort==fcall_es && expr->decl && expr->decl->inst)
  {
    return(mksubst(expr->decl->inst));
  }

      /* If _._ is encountered, blindly copy the first argument to
         stop it from being treated as an implicit parameter */
  
  if (expr->sort==fcall_es &&  expr->u.fc.ident==dot_ident)
  {
    p=palloc(sizeof(struct expr));
    *p = *expr;
    p->u.fc.args=mk_list(arg(1,expr),mk_list(mksubst(arg(2,expr)),NULL));
    p->idcode=idcode++;
    return p;
  }
  
      /* Substitute the prototype of an implicit parameter */

  if (expr->sort==fcall_es && expr->u.fc.ident==decl2_ident)
  {
    return(mksubst(arg(1,expr)));
  }
  
      /* Otherwise don't substitute anything; copy the formal. */
  
  p=palloc(sizeof(struct expr));
  *p = *expr;
  if (expr->sort==fcall_es)
    p->u.fc.args=mksubst_args(expr->u.fc.args);
  p->idcode=idcode++;
  return(p);
}


struct list * mksubst_args ( struct list * args )
{
  if (args)
      return(mk_list(mksubst(args->expr),
		     mksubst_args(args->next)));
  else
      return(NULL);
}
