/*
    The Language for Asynchronous Research and Development -- LARD
    Copyright (C) 1996-1999 AMULET Group, Department of Computer Science
        The University of Manchester, Oxford Road, Manchester, UK, M13 9PL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Id: infix.c,v 1.1.1.1 1999/07/15 18:47:33 bardslea Exp $


	Infix expression parser

	11 Jun 1999, AMULET Group
*/

#include <string.h>
#include <time.h>

#include <alloc.h>
#include <error.h>
#include <simple.h>

#include "debug.h"
#include "parser_structs.h"
#include "dump.h"
#include "infix.h"
#include "codegen.h"

bool debug_pass_1d_more=FALSE;
int callocs;
int cfrees;
int iallocs;
int ifrees;
bool recursed=FALSE;

#ifndef CLOCKS_PER_SEC
/* Assume microseconds */
#define CLOCKS_PER_SEC (1000000)
#endif

/* This is the LARD infix expression parser.  Its role is to convert
   an infix expression, such as a*b+c*d, into the equivlanet prefix
   expression, in this case _+_(_*_(a,b),_*_(c,d)).

   The input infix expression is a linked list, each element of which
   is either a keyword or a subexpression.  The above example would be
   represented by the list [a,*,b,+,c,*,d], where a, b, c and d are
   subexpressions and * and + are keywords.  The types for the infix
   expression are defined in parser_structs.h.

   The output prefix expression is in the standard compiler format,
   also defined in parser_structs.h.

   The interface functions are:
     add_rule: add a new rule to the grammar
     infix_to_expr: parse an infix expression according to the current
                    rules and return an equivalent prefix expr.

   Each rule coresponds to an infix declaration in the input, such as
   _+_ : infix(160,left).  add_rule adds this information to a list of
   rules.  firstrule is a pointer to the start of this list, which is
   sorted in increasing order of priority.  Each list element is a
   struct rule, which contains header information (priority,
   associativity, prefix identifier) and a pointer to a linked list of
   struct rulebits.  The list of rulebits is similar to an infix
   expression; each bit is either a keyword or a placeholder for a
   subexpression (coresponding to an _ in the prefix identifier).

   The crucial thing about this problem is that the rules are defined
   dynamically, so a bison style parser is impossible.

   The grammar can be considered as follows:

   The start symbol is the "lowest priority expression", E0.  Then
   initially E0::=E1, E1::=E2, E2::=E3, E3::=E4 .... Einf-1::=Einf.
   Also, Einf::=subexpression.  So in this initial condition the only
   possible production leads to subexpression, which is a terminal
   symbol.

   Rules are then added to this grammar.  For example the above rule
   for _+_ defines a rule E160 ::= E160 + E161 where + is a terminal
   symbol.  Note that because _+_ is left associative (i.e. a+b+c ==
   (a+b)+c), the nonterminal to the right of the + is E161 whereas the
   one to the left is E160.  This allows a further _+_ for the left
   operand but not for the right.  This is reversed for right
   associative operators such as _;_.

   The parser is a top-down recursive decent parser.  For an
   introduction to this type of parser please see "Introduction to
   Computer Theory" by Daniel Cohen, chapter 22, particularly pages
   504 to 510.

   The parsing process operates as follows:

   We start with a "candidate expression" containing the single
   nonterminal symbol E0.  We then generate further candidate
   expressions in a recursive fashion by applying all possible
   sequences of rule applications.  This exploration terminates when
   we have either (a) found the target expression, or (b) found an
   expression which cannot possibly become the target expression by
   any further modification.  If all branches of the exploration lead
   to outcome (b) then the original expression cannot be parsed.

   This recursive exploration is clearly potentially complex, so an
   initial optimisation is made.  The list of rules is scanned and
   only those that contain only keywords found in the target
   expression are retained.  This preprocessing is carried out by
   get_rule_subset and the sub_next field of struct rule is used to
   link the possible subset for the current target.

   The recursive expansion is implemented by topdownparse.  All
   possible productions of the left most nonterminal symbol in the
   current candidate expression are considered.  The nonterminal
   records the rule that produced it in its rule field.  It also
   records whether it was a left or right operand of that rule in its
   rule_pos field.  The struct rule maintains pointers to the first
   rule with the same priority and the next rule with higher priority,
   so topdownparse can find the first possible rule that can be
   applied to that nonterminal.  Applying higher priority rules is
   equivalent to first applying a sequence of En -> En+1 rules.

   The elements of the candidate expression are joined by two sets of
   linked-list pointers, next and rnext.  Consider the candidate
   expression E + E.  If the rule for _*_ is applied to this the
   result, following the next pointers, is E * E + E.  However the
   rnext pointers are not updated and continue to follow the original
   E + E expression.

   Function apply applies rule r to the candidate expression at the
   position of the left most nonterminal (provided).

   Having applied the rule the new candidate expression is compared
   with the target by function match.  An exact match indicates sucess
   and termination of the recursion.

   If an exact match is not found might_match is called to investigate
   whether this candidate expression could ever match after any
   further sequence of productions.  The key here is that once a
   terminal (keyword) has appeared in the candidate expression it will
   never go away, and furthermore that each nonterminal in the
   candidate must eventually be replaced by at least one terminal.  If
   the candidate can never match it is rejected and the next rule is
   considered.

   If the candidate might match but doesn't exactly match topdownparse
   is called recursively to investigate further.

   After each of the explicit rules has been applied the implicit rule
   Einf ::= subexpression is applied and considered in the same way.

   The final step is to extract the prefix expression.  This is done
   using the rnext and rule pointers to construct the prefix
   expression bottom-up using ctoe.

   */


    /* Types representing rules */

enum rulebitsort { kw_rs, expr_rs };

struct rule
{
  enum assocdirection assoc;
  int priority;
  struct rulebit * bits;
  struct symrec * ident;
  struct rule * next;        /* next rule in the permanent list */
  struct rule * sub_next;    /* next rule in the sub-list for this target */
  struct rule * first_same_pri;   /* first rule with the same priority
                                     as this one (for the sublist) */
  struct rule * first_next_pri;   /* first rule with a higher priority than
                                     this one (for the sublist) */
  bool uniq_kws;             /* does this rule contain only keywords that are
                                uniq to it in the current sublist */
};

struct rulebit
{
  enum rulebitsort sort;
  struct symrec * keyword;
  struct rulebit * next;
  bool higher_pri_only;      /* associativity - e.g. set for rhs of _+_ */
  int kwcount;               /* count of occurences of this keyword in the
                                current sublist (valid if samekw is NULL) */
  struct rulebit * samekw;   /* pointer to another bit with the same
                                keyword */
};


    /* Types representing candidate expressions */

enum candsort { kw_cs, texpr_cs, nexpr_cs };
    /* keyword, terminal subexpression, nonterminal subexpression */

struct cand
{
  enum candsort sort;
  struct symrec * keyword;
  struct cand * param;      /* for a nonterminal subexpression which
                               has been expanded by the application
                               of a rule, this points to the start of
                               the expansion. */
  struct rule * rule;       /* for a nonterminal subexpression which
                               has been expanded by the application of
                               a rule, this points to the rule used
                               for the expansion. */
  struct cand * next;       /* next bit of the candidate expression.
                               if the next bit was a nonterminal
                               subexpression which has since been
                               expanded, this points to the expanded
                               version.  */
  struct cand * rnext;      /* next bit of the candidate expression.
                               if the next bit was a nonterminal
                               subexpression which has since been
                               expanded, this points to the
                               *un*expanded version.  */
  struct infix * iptr;      /* if the candidate expression matched the
                               target infix expression, this points to
                               the bit of the infix expression
                               coresponding to this bit of the
                               candidate. */
};


void add_rule ( char * name, enum assocdirection assoc,
                int priority );
void add_rule_i ( struct rule * newrule, struct rule * * r );
struct rulebit * make_bits ( char * name );
struct expr * infix_to_expr ( struct infix * infix );
struct cand * initialcandidate ( struct rule * rule );
bool topdownparse ( struct infix * target, struct cand * candidate );
void apply ( struct rule * r, struct cand * * c, struct cand * e );
void apply_bits ( struct rule * rule, struct rulebit * r,
                  struct cand * * c, struct cand * e );
bool match ( struct cand * candidate, struct infix * target );
bool might_match ( struct cand * candidate, struct infix * target );
bool might_match_tail ( struct cand * candidate, struct infix * target );
struct expr * cand_to_expr ( struct cand * c );
struct list * cand_to_expr_args ( struct cand * c, struct cand * stop );
struct rule * get_rule_subset ( struct rule * rule, struct infix * infix );
void find_kwuniq_rules ( struct rule * rules );
void count_keywords ( struct rule * rules );
void count_keywords_bits ( struct rulebit * rb, struct rule * next_rule );
int do_kwcount ( struct rulebit * firstbit, struct symrec * kw,
                 struct rulebit * rb, struct rule * next_rule );
void mark_uniq ( struct rule * rules );
bool uniq_keywords ( struct rulebit * rb );
bool possible_rule ( struct rulebit * rb, struct infix * infix );
struct rule * set_pri_pointers ( struct rule * rule,
                                 struct rule * first_same );
bool operatorparse ( struct infix * * infixp, struct rule * rules,
                     int prev_priority );
void op_apply_left ( struct rule * rule, struct infix * * infixp );
void op_apply_right ( struct rule * rule, struct infix * * infixp );
bool op_match ( struct rulebit * rb, struct infix * infix );
void op_substitute ( struct infix * * infixp, struct rule * rule );
struct list * op_get_args ( struct infix * infix, struct rulebit * rulebit );
struct infix * op_get_remainder ( struct infix * infix,
                                  struct rulebit * rulebit );
struct expr *  bup_pred_parse ( struct infix * infix, struct rule * rules );
struct list * bup_apply_right ( struct infix * infix, struct rulebit * rb );
struct infix * bup_right_scan ( struct infix * infix,
                                struct symrec * keyword );
void free_infix ( struct infix * i );
void dump_infix ( struct infix * infix );
void dump_cand ( struct cand * cand );
void dump_rulebits ( struct rulebit * rulebit );
void dump_rule ( struct rule * rule );


    /* The rules */

struct rule * firstrule=NULL;


    /* Functions to add a new rule to the grammar */

void add_rule ( char * name, enum assocdirection assoc,
                int priority )
{
  struct rule * newrule;

  debug(debug_pass_1d,"Adding rule for %s with priority %d and direction %s\n",
        name,priority,
        assoc==left_assoc?"left":assoc==right_assoc?"right":
        assoc==both_assoc?"both":"none");

  if (gen_dcode)
  {
    fputc('I',dcode_file);
    fputs(name,dcode_file);
    fputc('\0',dcode_file);
    fwrite(&priority,sizeof(int),1,dcode_file);
    fputc(assoc==left_assoc?'L':assoc==right_assoc?'R':
          assoc==both_assoc?'B':'N',dcode_file);
  }
  
  newrule=palloc(sizeof(struct rule));
  newrule->assoc=assoc;
  newrule->priority=priority;
  newrule->bits=make_bits(name);
  newrule->ident=putsym(name);

  if (assoc==left_assoc)
  {
    struct rulebit * rb;

    rb=newrule->bits;
    while (rb->next)
      rb=rb->next;
    rb->higher_pri_only=TRUE;
  } else if (assoc==right_assoc)
  {
    newrule->bits->higher_pri_only=TRUE;
  } else if (assoc==both_assoc)  {
    struct rulebit * rb;

    rb=newrule->bits;
    while (rb)
    {
      rb->higher_pri_only=TRUE;
      rb=rb->next;
    }
  }
  
  add_rule_i(newrule,&firstrule);
}

void add_rule_i ( struct rule * newrule, struct rule * * r )
{
  if (!(*r))
  {
    newrule->next=NULL;
    (*r)=newrule;
    return;
  }
  
  if (newrule->priority<(*r)->priority)
  {
    newrule->next=(*r);
    (*r)=newrule;
  } else
  {
    add_rule_i(newrule,&((*r)->next));
  }
}


struct rulebit * make_bits ( char * name )
{
  struct rulebit * rb;
  char * ul;
  
  if ((*name)=='\0')
  {
    return NULL;
  }

  rb=palloc(sizeof(struct rulebit));

  if ((*name)=='_')
  {
    rb->sort=expr_rs;
    rb->next=make_bits(&name[1]);
    rb->higher_pri_only=FALSE;
  } else
  {
    rb->sort=kw_rs;
    ul=strchr(name,'_');
    if (ul)
    {
      (*ul)='\0';
      rb->keyword=putsym(name);
      (*ul)='_';
      rb->next=make_bits(ul);
    } else
    {
      rb->keyword=putsym(name);
      rb->next=NULL;
    }
    rb->keyword->sort=iskeyword;
  }

  return(rb);
}


    /* Parsing functions */

struct expr * infix_to_expr ( struct infix * infix )
{
  struct rule * rules;
  struct cand * match;
  struct expr * e;
  struct infix * i;
  int start;
  bool oldrecursed;
  
  iallocs=0;
  i=infix;
  while (i)
  {
    iallocs++;
    i=i->next;
  }
  
  if (debug_pass_1d && !recursed)
  {
    fprintf(stderr,"Processing ");
    dump_infix(infix);
    fprintf(stderr,"\n");
    callocs=0;
    cfrees=0;
    ifrees=0;
  }

  start=clock();
  
  rules=get_rule_subset(firstrule,infix);

  if (!rules)
  {
    nonfatal_error(infix->line,"Can't decode infix expression ");
    dump_infix(infix);
    fprintf(stderr," (no rules match)\n");
    exit(1);
  }
  
  find_kwuniq_rules(rules);
  oldrecursed=recursed;
/*  recursed=TRUE; */
  e=bup_pred_parse(infix,rules);
  recursed=oldrecursed;
  if (e)
  {
    if (debug_pass_1d && !recursed)
    {
      fprintf(stderr,"Terminating after bottom-up predictive parsing,"
              "took %0.2fms, got ",
              (double)1000.0*(clock()-start)/CLOCKS_PER_SEC);
      dump_expr(e,-1);
      fprintf(stderr,"\n");
    }
    return e;
  }
  
  operatorparse(&infix,rules,-1);

  if (debug_pass_1d_more)
  {
    fprintf(stderr,"After operator parsing: ");
    dump_infix(infix);
    fprintf(stderr,"\n");
  }

  if (!infix->next && infix->sort==expr_is)
  {
    e=infix->u.expr;
    free_infix(infix);
    if (debug_pass_1d && !recursed)
    {
      fprintf(stderr,"Terminating after operator parsing, "
              "took %0.2fms, got ",
              (double)1000.0*(clock()-start)/CLOCKS_PER_SEC);
      dump_expr(e,-1);
      fprintf(stderr,"\n");
    }
    return e;
  }
  
  rules=get_rule_subset(firstrule,infix);
  set_pri_pointers(rules,rules);
  match=initialcandidate(rules);
  if (!topdownparse(infix,match))
  {
    nonfatal_error(infix->line,"Can't decode infix expression ");
    dump_infix(infix);
    fprintf(stderr,"\n");
    exit(1);
  }

  e=cand_to_expr(match);
  free(match);
  cfrees++;
  free_infix(infix);
  
  if (debug_pass_1d && !recursed)
  {
    fprintf(stderr,"done, "
            "took %0.2fms, got ",
            (double)1000.0*(clock()-start)/CLOCKS_PER_SEC);
    dump_expr(e,-1);
    fprintf(stderr,"\n");
  }

  return e;
}


/* Return a candidate expression containing a single nonterminal. */

struct cand * initialcandidate ( struct rule * rule )
{
  struct cand * c;
  callocs++;
  c=palloc(sizeof(struct cand));
  c->sort=nexpr_cs;
  c->next=NULL;
  c->rnext=NULL;
  c->rule=rule;
  return c;
}


    /* recursively expand the candidate expression until either an
       expression matching the target is found (return true) or all
       recursive expansions have been tried (return false). */

bool topdownparse ( struct infix * target, struct cand * candidate )
{
  struct rule * r;
  struct cand * lmnt;
  struct cand * * lmntp;
  struct cand * e;
  struct cand * p;
  struct cand * fp;
  struct cand * fp2;
  
  
      /* Find the leftmost nonterminal subexpression in the candidate
         expression.  lmntp is a pointer to the pointer to it in the
         bit to its left */
  
  lmntp=&candidate;
  while ((*lmntp) && (*lmntp)->sort!=nexpr_cs)
  {
    lmntp=&(*lmntp)->next;
  }
  lmnt=(*lmntp);

      /* If there is no nonterminal, this candidate is a dead-end. */
  
  if (!lmnt)
  {
    if (debug_pass_1d_more)
    {
      fprintf(stderr,"candidate contains no non-terminals\n");
    }
    return FALSE;
  }

  if (debug_pass_1d_more)
  {
    fprintf(stderr,"topdownparse: current candidate = ");
    dump_cand(candidate);
    fprintf(stderr,"\n");
  }

      /* We are going to consider each rule in turn by pasting in a
         new section of candidate expression in place of the leftmost
         nonterminal.  e is a pointer to the bit after the
         nonterminal, used for joining it on. */
  
  e=lmnt->next;

      /* We start with the rule pointer to by the leftmost
         nonterminal.  For the initial candidate expression this is
         the first rule in the sublist.  Subsequently it will have
         been filled in as appropriate so that we process rules with
         priority >= or > than the priority of the rule that created
         this nonterminal. */
  
  r=lmnt->rule;
  
  while (r)
  {
    lmnt->rule=r;           /* record that this nonterminal was
                               expanded by this rule */
    apply(r,lmntp,e);       /* replace this nonterminal by a new
                               section of candidate expression
                               coresponding to the rule */
    fp=(*lmntp); /* for freeing if this expansion fails */
    lmnt->param=(*lmntp);   /* make the nonterminal point to the start
                               of its expansion, which coresponds to
                               its parameters in the prefix expression */
    if (debug_pass_1d_more)
    {
      fprintf(stderr,"trying rule E -> ");
      dump_rule(r);
      fprintf(stderr,", new candidate is ");
      dump_cand(candidate);
      fprintf(stderr,"\n");
    }
    if (match(candidate,target))    /* return TRUE if we have a
                                       perfect match */
    {
      if (debug_pass_1d_more)
      {
        fprintf(stderr,"matches!\n");
      }
      return TRUE;
    } else if (might_match(candidate,target))
    {
      if (debug_pass_1d_more)
      {
        fprintf(stderr,"might match, descending\n");
      }
      if (topdownparse(target,candidate))  /* if the expansion might
           (eventually) match the target we continue the processing
           recursively, i.e. we apply all appropriate rules to the
           next leftmost nonterminal (which will either be the
           leftmost nonterminal of the expansion we have just done, or
           if that contained only terminals, the next nonterminal in
           the original candidate expression). */
      {
        return TRUE;  /* If the recursive call is sucessful, so are we */
      }
    } else
    {
      if (debug_pass_1d_more)
      {
        fprintf(stderr,"rejected\n");
      }
    }
        /* Free the failed expansion, i.e. from (*lmntp) up to e */
    while (fp!=e)
    {
      fp2=fp->rnext;
      cfrees++;
      free(fp);
      fp=fp2;
    }
    (*lmntp)=lmnt;
    r=r->sub_next;     /* Otherwise we consider the next rule */
  }
      /* At this point we have tried applying all the "normal" rules
         to the leftmost nonterminal without sucess.  Now we try the
         special rule that turns a nonterminal subexpression into a
         terminal subexpression.  This code is similar to the general
         case above.*/

  lmnt->rule=NULL;                 /* special marker for this rule */
  callocs++;
  p=palloc(sizeof(struct cand));
  p->sort=texpr_cs;
  p->next=e;
  p->rnext=e;
  (*lmntp)=p;
  lmnt->param=p;
  if (debug_pass_1d_more)
  {
    fprintf(stderr,"trying rule E -> (E) , new candidate is ");
    dump_cand(candidate);
    fprintf(stderr,"\n");
  }
  if (match(candidate,target))
  {
    if (debug_pass_1d_more)
    {
      fprintf(stderr,"matches!\n");
    }
    return TRUE;
  } else if (might_match(candidate,target))
  {
    if (debug_pass_1d_more)
    {
      fprintf(stderr,"might match, descending\n");
    }
    if (topdownparse(target,candidate))
    {
      return TRUE;
    }
  } else
  {
    if (debug_pass_1d_more)
    {
      fprintf(stderr,"rejected\n");
    }
  }
  free(p);
  cfrees++;
  (*lmntp)=lmnt;
      /* If we get here we have tried all expansions from the
              current candidate expansion (note: all expansion of the
              whole expression, not just expansions of the leftmost
              nonterminal), and none has matched, so we give up.  The
              caller will then find another candidate expression and
              try again. */
  if (debug_pass_1d_more)
  {
    fprintf(stderr,"run out of rules, returning\n");
  }
  return FALSE;
}


    /* Add expansion of rule r to candidate expression at (*c)...e */

void apply ( struct rule * r, struct cand * * c, struct cand * e )
{
  apply_bits(r,r->bits,c,e);
}


void apply_bits ( struct rule * rule, struct rulebit * r,
                  struct cand * * c, struct cand * e )
{
  struct cand * p;

  if (!r)
  {
    (*c)=e;
  } else
  {
    callocs++;
    p=palloc(sizeof(struct cand));
    p->sort=(r->sort==kw_rs?kw_cs:nexpr_cs);
    if (r->sort==kw_rs) p->keyword=r->keyword;
    apply_bits(rule,r->next,&p->next,e);
    p->rnext=p->next;
    if (r->sort==expr_rs && r->higher_pri_only)
    {
      p->rule=rule->first_next_pri;
    } else {
      p->rule=rule->first_same_pri;
    }
    (*c)=p;
  }
}


    /* Test if candidate expression exactly matches target */

bool match ( struct cand * candidate, struct infix * target )
{
  if (!candidate && !target)
    return TRUE; /* end of expression reached */

  if (!target || !candidate)
    return FALSE; /* lengths differ */

  candidate->iptr=target;

  if (candidate->sort==kw_cs && target->sort==kw_is)
  {
    if (candidate->keyword==target->u.keyword)
    {
      return match(candidate->next,target->next);
    } else
    {
      return FALSE; /* keywords mismatch */
    }
  } else if (candidate->sort==kw_cs && target->sort==expr_is)
  {
    return FALSE; /* subexpression expected but keyword found */
  } else if (candidate->sort==nexpr_cs)
  {
    return FALSE; /* target can't contain noterminals */
  } else if (candidate->sort==texpr_cs && target->sort==kw_is)
  {
    return FALSE; /* keyword expected but subexpression found */
  } else if (candidate->sort==texpr_cs && target->sort==expr_is)
  {
    return match(candidate->next,target->next);
  }
  internal_error("shouldn't be here (a)");
  return FALSE;
}


    /* Test if candidate expression might match target after further
       expansion.  Look for an exact match up to the first nonterminal
       in the candidate. */

bool might_match ( struct cand * candidate, struct infix * target )
{
  if (!candidate && !target)
    internal_error("shouldn't be here (b)"); /* exact match! */

  if (!target || !candidate)
    return FALSE; /* exact match so far but lengths differ */

  if (candidate->sort==kw_cs && target->sort==kw_is)
  {
    if (candidate->keyword==target->u.keyword)
    {
      return might_match(candidate->next,target->next);
    } else
    {
      return FALSE; /* keywords in initial substring mismatch */
    }
  } else if (candidate->sort==kw_cs && target->sort==expr_is)
  {
    return FALSE; /* subexpression expected but keyword found */
  } else if (candidate->sort==nexpr_cs)
  {
    return might_match_tail(candidate,target);
  } else if (candidate->sort==texpr_cs && target->sort==kw_is)
  {
    return FALSE; /* keyword expected but subexpression found */
  } else if (candidate->sort==texpr_cs && target->sort==expr_is)
  {
    return might_match(candidate->next,target->next);
  }
  internal_error("shouldn't be here (c)");
  return FALSE;
}


    /* After the first nonterminal come here for less precise checking */

    /* *** Could also reject more cases by checking for right
       substrings *** */

bool might_match_tail ( struct cand * candidate, struct infix * target )
{
  if (!candidate)
    return TRUE;   /* run out of candidate, but it might have ended in
                      a nonterminal which would have expanded to match
                      the rest of the target */

  if (!target)
    return FALSE;  /* run out of target, so remaining elements of the
                      candidate have nothing to match */

  if (candidate->sort==kw_cs && target->sort==kw_is)
  {
    if (candidate->keyword==target->u.keyword)
    {
          /* keywords match, keep going */
      return might_match_tail(candidate->next,target->next);
    } else
    {
          /* keywords don't match, advance through target until
             candidate keyword is found */
      return might_match_tail(candidate,target->next);
    }
  } else if (candidate->sort==kw_cs && target->sort==expr_is)
  {
        /* advance through target until candidate keyword is found */
    return might_match_tail(candidate,target->next);
  } else if (candidate->sort==nexpr_cs)
  {
        /* nonterminal expression in candidate matches either keyword
           or expression in target */
    return might_match_tail(candidate->next,target->next);
  } else if (candidate->sort==texpr_cs && target->sort==kw_is)
  {
        /* looking for an expression but got a keyword, advance
           through target until an expression is found */
    return might_match_tail(candidate,target->next);
  } else if (candidate->sort==texpr_cs && target->sort==expr_is)
  {
        /* expressions match, keep going */
    return might_match_tail(candidate->next,target->next);
  }
  fprintf(stderr,"cand is %s and target is %s\n",
          candidate->sort==kw_cs?"kw":
          candidate->sort==nexpr_cs?"nexpr":
          candidate->sort==texpr_cs?"texpr":"UDEF",
          target->sort==kw_is?"kw":
          target->sort==expr_is?"expr":"UDEF");
  internal_error("shouldn't be here (d)");
  return FALSE;
}


    /* Construct the prefix expression from the final matching
       candidate expression */

struct expr * cand_to_expr ( struct cand * c )
{
  struct expr * e;
  if (!c->rule)  /* terminal subexpression */
  {
    free(c->param);
    cfrees++;
    e=c->iptr->u.expr;  /* iptr filled in by match */
  } else {
    e=mk_expr(c->rule->ident,cand_to_expr_args(c->param,c->rnext));
  }
  return e;
}


struct list * cand_to_expr_args ( struct cand * c, struct cand * stop )
{
  struct list * l;

  if (c==stop)
    return NULL;
  
  if (c->sort==nexpr_cs)
  {
    l=mk_list(cand_to_expr(c),cand_to_expr_args(c->rnext,stop));
  } else
  {
    l=cand_to_expr_args(c->rnext,stop);
  }
  free(c);
  cfrees++;
  return l;
}


    /* Rule subset creation functions */

struct rule * get_rule_subset ( struct rule * rule, struct infix * infix )
{
  if (!rule)
  {
    return NULL;
  }
  
  if (possible_rule(rule->bits,infix))
  {
    if (debug_pass_1d_more)
    {
      dump_rule(rule);
      fprintf(stderr," is a candidate\n");
    }
    rule->sub_next=get_rule_subset(rule->next,infix);
    return rule;
  } else
  {
    return get_rule_subset(rule->next,infix);
  }
}


bool possible_rule ( struct rulebit * rb, struct infix * infix )
{
  if (!rb)
    return TRUE;

  if (!infix)
    return FALSE;
  
  rb->samekw=NULL;

  if (rb->sort==kw_rs)
  {
    if (infix->sort==kw_is)
    {
      if (rb->keyword==infix->u.keyword)
      {
        return possible_rule(rb->next,infix->next);
      } else
      {
        return possible_rule(rb,infix->next);
      }
    } else
    {
      return possible_rule(rb,infix->next);
    }
  } else
  {
    if (infix->sort==kw_is)
    {
      return possible_rule(rb,infix->next);
    } else
    {
      return possible_rule(rb->next,infix->next);
    }
  }
}


    /* Scan the list of possible rules and find those that contain
       only keywords that no other rule contains */

void find_kwuniq_rules ( struct rule * rules )
{
  count_keywords(rules);
  mark_uniq(rules);
}

void count_keywords ( struct rule * rules )
{
  if (rules)
  {
    count_keywords_bits(rules->bits,rules->sub_next);
    count_keywords(rules->sub_next);
  }
}

void count_keywords_bits ( struct rulebit * rb, struct rule * next_rule )
{
  if (rb)
  {
    if (rb->sort==kw_rs)
    {
      if (rb->samekw)
      {
            /* This bit contains a keyword that has already been counted */
        rb->kwcount=rb->samekw->kwcount;
      } else
      {
            /* This bit contains a keyword that hasn't been counted yet */
        rb->kwcount=1+do_kwcount(rb,rb->keyword,rb->next,next_rule);
        if (debug_pass_1d_more)
        {
          fprintf(stderr,"%d occurances of %s\n",rb->kwcount,
                  rb->keyword->name);
        }
      }
    }
    count_keywords_bits(rb->next,next_rule);
  }
}


int do_kwcount ( struct rulebit * firstbit, struct symrec * kw,
                 struct rulebit * rb, struct rule * next_rule )
{
  if (!rb)
  {
    if (!next_rule)
    {
      return 0;
    } else
    {
      return do_kwcount(firstbit,kw,next_rule->bits,next_rule->sub_next);
    }
  } else
  {
    if (rb->sort==kw_rs && rb->keyword==kw)
    {
      rb->samekw=firstbit;
      return 1+do_kwcount(firstbit,kw,rb->next,next_rule);
    } else
    {
      return do_kwcount(firstbit,kw,rb->next,next_rule);
    }
  }
}


void mark_uniq ( struct rule * rules )
{
  if (rules)
  {
    rules->uniq_kws=uniq_keywords(rules->bits);
    if (debug_pass_1d_more)
    {
      if (rules->uniq_kws)
      {
        dump_rule(rules);
        fprintf(stderr," contains only unique keywords\n");
      }
    }
    mark_uniq(rules->sub_next);
  }
}


bool uniq_keywords ( struct rulebit * rb )
{
  if (rb)
  {
    if (rb->sort==kw_rs)
    {
      if (rb->samekw && rb->samekw->kwcount==1)
      {
        internal_error("not likely!!!");
        return uniq_keywords(rb->next);
      } else if (!rb->samekw && rb->kwcount==1)
      {
        return uniq_keywords(rb->next);
      } else
      {
        return FALSE;
      }
    } else
    {
      return uniq_keywords(rb->next);
    }
  } else
  {
    return TRUE;
  }
}



    /* Set up the first_same_pri and first_next_pri pointers in the
       rule sublist */

struct rule * set_pri_pointers ( struct rule * rule, struct rule * first_same )
{
  struct rule * rcr;
  
  if (rule)
  {
    if (rule->priority==first_same->priority)
    {
      rule->first_same_pri=first_same;
      rcr=set_pri_pointers(rule->sub_next,first_same);
    } else
    {
      rule->first_same_pri=rule;
      rcr=set_pri_pointers(rule->sub_next,rule);
    }
    if (!rule->sub_next)
    {
      rule->first_next_pri=NULL;
    } else if (rule->priority==rule->sub_next->priority)
    {
      rule->first_next_pri=rcr;
    } else
    {
      rule->first_next_pri=rule->sub_next;
    }
    return rule->first_next_pri;
  } else
  {
    return NULL;
  }
}


    /* The operator-parser */

bool operatorparse ( struct infix * * infixp, struct rule * rules,
                     int prev_priority )
{
      /* Rules are ordered in increasing priority, and we want to
         apply them in decreasing priority, so we do the recursive
         call first */

      /* We have to stop when we get to a rule that contains
         non-unique keywords */

      /* We also have to stop when we get to a rules that has the same
         priority as another rule, so we pass down the priority of the
         current rule in the recursive call.  If this value on entry
         is the same as the current rule's priority we give up. */

      /* The recursive call returns false if this has already happened */

  if (!rules)
    return TRUE;

  if (!operatorparse(infixp,rules->sub_next,rules->priority))
    return FALSE;

  if (rules->priority==prev_priority)
    return FALSE;
  
  if (!rules->uniq_kws)
    return FALSE;

      /* Now we can get on and apply this rule */

  if (debug_pass_1d_more)
  {
    fprintf(stderr,"operatorparse: applying ");
    dump_rule(rules);
    fprintf(stderr,"\n");
  }

  if (rules->assoc==left_assoc)
  {
    op_apply_left(rules,infixp);
  } else
  {
    op_apply_right(rules,infixp);
  }
  return TRUE;
}


void op_apply_left ( struct rule * rule, struct infix * * infixp )
{
  if (!(*infixp))
    return;

  if (op_match(rule->bits,(*infixp)))
  {
    if (debug_pass_1d_more)
    {
      dump_rule(rule);
      fprintf(stderr," matched\n");
    }
    op_substitute(infixp,rule);
    op_apply_left(rule,infixp);
  } else
  {
    op_apply_left(rule,&((*infixp)->next));
  }
}


void op_apply_right ( struct rule * rule, struct infix * * infixp )
{
  if (!(*infixp))
    return;

  op_apply_right(rule,&((*infixp)->next));
  if (op_match(rule->bits,(*infixp)))
  {
    if (debug_pass_1d_more)
    {
      dump_rule(rule);
      fprintf(stderr," matched\n");
    }
    op_substitute(infixp,rule);
  }
}


bool op_match ( struct rulebit * rb, struct infix * infix )
{
  if (!rb)
    return TRUE;

  if (!infix)
    return FALSE;

  if ((rb->sort==expr_rs && infix->sort==expr_is)
      || (rb->sort==kw_rs && infix->sort==kw_is
          && rb->keyword==infix->u.keyword))
    return op_match(rb->next,infix->next);
  else
    return FALSE;
}


void op_substitute ( struct infix * * infixp, struct rule * rule )
{
  struct list * l;
  struct expr * e;
  struct infix * i;
  struct infix * rest;

  l=op_get_args((*infixp),rule->bits);
  e=mk_expr(rule->ident,l);
  rest=op_get_remainder((*infixp),rule->bits);
  i=mk_infix_expr(e,rest);
  iallocs++;
  (*infixp)=i;
}


struct list * op_get_args ( struct infix * infix, struct rulebit * rulebit )
{
  if (!rulebit)
    return NULL;

  if (rulebit->sort==expr_rs)
  {
    return mk_list(infix->u.expr,op_get_args(infix->next,rulebit->next));
  } else
  {
    return op_get_args(infix->next,rulebit->next);
  }
}


struct infix * op_get_remainder ( struct infix * infix,
                                  struct rulebit * rulebit )
{
  struct infix * inext;
  
  if (!rulebit)
    return infix;

  inext=infix->next;
  free(infix);
  ifrees++;
  return op_get_remainder(inext,rulebit->next);
}


    /* The bottom-up predictive parser */

struct expr *  bup_pred_parse ( struct infix * infix, struct rule * rule )
{
      /* Can't do anything if the rule doesn't have only unique keywords */
  if (!rule->uniq_kws)
    return NULL;

      /* Can't do anything if there are other rules with the same priority */
  if (rule->sub_next && rule->priority==rule->sub_next->priority)
    return NULL;

      /* Can only parse right assoc rules at the moment */
  if (rule->assoc==left_assoc)
    return NULL;

  if (debug_pass_1d_more)
  {
    fprintf(stderr,"bup_pred_parse: applying ");
    dump_rule(rule);
    fprintf(stderr,"\n");
  }

  return mk_expr(rule->ident,bup_apply_right(infix,rule->bits)); /* *** */
}


struct list * bup_apply_right ( struct infix * infix, struct rulebit * rb )
{
  struct expr * e;
  struct list * l;
  struct infix * rest;
  struct infix * n;
  
  if (!infix)
    return NULL;
  
  if (rb->sort==kw_rs && infix->sort==kw_is)
  {
    if (rb->keyword==infix->u.keyword)
    {
      n=infix->next;
      free(infix);
      ifrees++;
      return bup_apply_right(n,rb->next);
    } else {
      internal_error("keyword mismatch");
    }
  } else if (rb->sort==kw_rs)
  {
    nonfatal_error(infix->line,"Can't decode infix expression ");
    dump_infix(infix);
    fprintf(stderr," (keyword expected but expression found)\n");
    exit(1);
  } else
  {
    if (!rb->next)
    {
      rest=NULL;
    } else if (rb->next->sort==expr_rs)
    {
      internal_error("adjacent expressions");
    } else
    {
      rest=bup_right_scan(infix,rb->next->keyword);
    }
    if (infix->sort==expr_is && !infix->next)
    {
      e=infix->u.expr;
      free(infix);
      ifrees++;
    } else
    {
      e=infix_to_expr(infix);
    }
    l=bup_apply_right(rest,rb->next);
    return mk_list(e,l);
  }
}


struct infix * bup_right_scan ( struct infix * infix, struct symrec * keyword )
{
  struct infix * n;

  if (infix->next->sort==kw_is)
  {
    if (infix->next->u.keyword==keyword)
    {
      n=infix->next;
      infix->next=NULL;
      return n;
    }
  }
  return bup_right_scan(infix->next,keyword);
}


    /* Misc */

void free_infix ( struct infix * i )
{
  if (i)
  {
    ifrees++;
    free_infix(i->next);
    free(i);
  }
}


    /* debugging functions */

void dump_infix ( struct infix * infix )
{
  if (!infix)
    return;

  if (infix->sort==kw_is)
    fprintf(stderr,"%s ",infix->u.keyword->name);

  else if (infix->sort==expr_is)
  {
    fprintf(stderr,"(");
    dump_expr(infix->u.expr,-1);
    fprintf(stderr,") ");
  }
  
  else
    fprintf(stderr,"ERROR ");
  
  dump_infix(infix->next);
}
      

void dump_cand ( struct cand * cand )
{
  if (!cand)
    return;

  if (cand->sort==kw_cs)
    fprintf(stderr,"%s ",cand->keyword->name);

  else if (cand->sort==texpr_cs)
    fprintf(stderr,"(E) ");

  else if (cand->sort==nexpr_cs)
    fprintf(stderr,"E ");

  else
    fprintf(stderr,"ERROR ");

  dump_cand(cand->next);
}
      

void dump_rulebits ( struct rulebit * rulebit )
{
  if (!rulebit)
    return;

  if (rulebit->sort==kw_rs)
    fprintf(stderr,"%s ",rulebit->keyword->name);
  else if (rulebit->sort==expr_rs)
    fprintf(stderr,"E ");
  else
    fprintf(stderr,"ERROR ");
  
  dump_rulebits(rulebit->next);
}


void dump_rule ( struct rule * rule )
{
  dump_rulebits(rule->bits);
}

