;;; lard-mode.el --- a major mode for LARD program source files

;;; Copyright (C) 1998 Arvid Grtting
;;
;; Author: Arvid Grtting <arvidg@ifi.uio.no>
;; Version: 0.1
;; Keywords:      LARD
;;
;;	$Id: lard-mode.el,v 1.1.1.1 1999/07/15 18:47:55 bardslea Exp $	
;;
;; Based loosely on mudela-mode.el, for which the following copyright
;; applies:
;;
;; Copyright (C) 1992,1993,1994  Tim Peters
;;
;; Author: 1997: Han-Wen Nienhuys
;; Author: 1995-1996 Barry A. Warsaw
;;         1992-1994 Tim Peters
;;
;; This software is provided as-is, without express or implied
;; warranty.  Permission to use, copy, modify, distribute or sell this
;; software, without fee, for any purpose and by any individual or
;; organization, is hereby granted, provided that the above copyright
;; notice and this paragraph appear in all copies.


;;; This file is NOT part of GNU Emacs.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;;; Commentary:
;;
;; This file enables a major mode for LARD: `lard-mode'.  LARD is the
;; Language for Asyncronous Research and Development developed at
;; the University of Manchester for the Amulet project; see
;; <URL: http://www.cs.man.ac.uk/amulet/projects/lard/>
;;
;; `lard-mode' defines a list of keywords and patterns for
;; fontification in `lard-font-lock-keywords' and tries to automate
;; indentation with `lard-indent-relative' and friends.  A hook for
;; customization is provided in `lard-mode-hook'.


;;; Installation:
;;
;; Place this file somewhere in your load-path and put something not
;; entirely unlike this in your .emacs:
;;
;; (autoload 'lard-mode "lard-mode" nil t)
;;
;; Please note that emacs usually assumes files with a file name
;; suffix of ".l" are lisp file; to change this assumption you could
;; put something like this in your .emacs:
;;
;; (setq auto-mode-alist
;;      (cons '("\\.l$" . lard-mode) auto-mode-alist))
;;
;; Personally, I want to auto-fill even Lard-mode buffers, and I want
;; to force them to use font-lock if I can, so I add this too:
;;
;; (add-hook 'lard-mode-hook
;;	  (lambda ()
;;	    (auto-fill-mode 1)
;;	    (if window-system
;;		(font-lock-mode 1))))

;;; Change log:
;;
;; 0.1  Initial revision

;;; TO DO:
;;
;; end-of-defun-like thing
;; functions to run lcd and li from the source code buffer

;; Ported to xemacs PBE 26/2/98; search for PBE for diffs

;; Modified by PAR 02/03/1998 to use one set of keybindings for Gnu and X
;;  search for PAR for diffs


;;; Code:

(defconst lard-font-lock-keywords

  (let* ((keywords '("use" "var" "expr" "type" "class" "int" "char"
		     "lib" "varargs" "overloaded" "bool" "scalar"
		     "string" "void" "true" "false" "null" "read"
		     "eval" "mkexpr" "mod" "and" "or" "xor" "not"
		     "record" "val_index_record" "var_index_record"
		     "succ" "pred" "ord"   "if" "then" "else" "while"
		     "do" "repeat" "until" "forever" "forseq" "in"
		     "to" "forpar" "forparnum" "wait" "exit" "atexit"
		     "case" "when" "default" "islib" "is" "rand"
		     "assert_version" "fatal" "sizeof_type" "val"
		     "sizeof_var" "sizeof_val" "peek" "poke" "cast"
		     "spawn" "running" "yield" "mytid" "priority"
		     "remote_waits" "setmypriority" "setpriority"
		     "foreach"))
	 (kwregex (mapconcat (lambda (x) 
			       (concat "" x))
			     keywords
			     "\\|")))
    (list
     ;; Keywords
     (cons (concat "[^a-zA-Z0-9_]\\(" kwregex "\\)[^a-zA-Z0-9_]") 1)
     (cons (concat "^\\(" kwregex "\\)[^a-zA-Z0-9_]") 1)
     (cons (concat "(\\(" kwregex "\\)[^a-zA-Z0-9_]") 1)
     ;;
     ;; include file names are strings, I guess (rather than variables)
     '("\\<use[ \t]*([ \t]*\\([a-zA-Z][a-zA-Z0-9_]*\\)[ \t]*)[ \t]*\\."
       1 font-lock-string-face)
     ;; classes and types
     '("\\b\\([a-zA-Z][a-zA-Z0-9_]*\\)[ \t]*:[ \t]*\\(class\\|type\\)\\b"
       1 font-lock-type-face)
     ;; functions
     '("\\b\\([a-zA-Z][a-zA-Z0-9_]*\\)[ \t]*\\((.*)\\)?[ \t]*:[ \t]*expr[ \t]*("
	  1 font-lock-function-name-face)
     ;; variable names
     '("\\<\\([a-zA-Z][a-zA-Z0-9_]*\\)[ \t]*:[ \t]*va[rl]\\>"
       1 font-lock-variable-name-face)
     '("\\<for\\(seq\\|each\\)[ \t]+\\([a-zA-Z][a-zA-Z0-9_]*\\)"
       2 font-lock-variable-name-face)
     ))
  "Additional expressions to highlight in Lard mode.
This includes keywords as well as function, variable and type names
(wherever the symbol is being declared).

Unfortunately, declarations where the symbol and its type are not on
the same line will not be highlit; this is, I think, due to the
line-by-line nature of font-lock.  If I'm mistaken, or if you find
other bugs, please tell me using the function `lard-mode-bug-report'")

;; define a mode-specific abbrev table for those who use such things
(defvar lard-mode-abbrev-table nil
  "Abbrev table in use in `lard-mode' buffers.")

(define-abbrev-table 'lard-mode-abbrev-table nil)

(defvar lard-mode-hook nil
  "Hook run when starting Lard-mode")

(defvar lard-indent-column 2
  "*Number of columns to indent each logical level of a Lard expression.")

(defun lard-mode-bug-report ()
  (interactive)
  "Send a bug report to the author of lard-mode"
  (condition-case err
      (message-mail "arvidg@ifi.uio.no" "lard-mode bug report")
    '(error (mail nil "arvidg@ifi.uio.no" "lard-mode bug report"))))

(defun lard-indent-relative ()
  "Indent current Lard line according to Lard syntax.
The line is indented by `lard-indent-column' columns for each level of
unclosed parantheses.  (Actually, this is acheived by looking on the
previous nonblank line's indentation as well as the number of opening
and closing parens between the beginning of that line and the
beginning of the current line.  If the first character of a line is a
closing paren, that line is indented one level less than the logical
level of the beginning of the line would indicate, i.e. to the logical
level outside this pair of parens."
  (interactive)
  (let (beg end indent someparen)
    (save-excursion
      (beginning-of-line)
      (setq end (point))
      (forward-line -1)
      (while (and (eolp) (< 0 (forward-line -1)))
					; do nothing
	)
      (setq beg (point))
      (skip-chars-forward " \t")
        (setq indent (current-column))
      ;; handle hanging closing parens
      (if (looking-at ")")
	  (setq indent (+ indent lard-indent-column)))
      (while (setq someparen (re-search-forward "[()]" end 'move-to-limit))
	(if (string-equal (match-string 0) "(")
	    (setq indent (+ indent lard-indent-column))
	  ;; it's a closing one
	  (setq indent (- indent lard-indent-column))))
      (if (< indent 0)
	  (progn
	    (message "Impossible indent column %s." indent)
	    (setq indent 0)))
      (indent-line-to indent)
      ;; handle hanging closing parens
      (if (looking-at ")")
	  (progn
	    (setq indent (if (< indent lard-indent-column) 0
			    (- indent lard-indent-column)))
	    (indent-line-to indent)))
      (move-to-column indent)
      (setq end (point)))
    ;; If we've ended up before the indentation, move to the end of it
    (if (> end (point))
	(goto-char end))))


(defun lard-electric-paren (arg)
  "Insert a parenthesis, then re-indent current line."
  (interactive "P")
  (self-insert-command (prefix-numeric-value arg))
  (lard-indent-relative))


(defun lard-newline-and-indent ()
  "Newline and indent in `lard-mode' buffer."
  (interactive)
  (newline)
  (lard-indent-relative))

(defvar lard-mode-map ()
  "Keymap for `lard-mode' buffers")

(if lard-mode-map
    ()
  (setq lard-mode-map (make-sparse-keymap))
  (mapcar (function (lambda (key)
		      (define-key
			lard-mode-map key 'lard-newline-and-indent)))
	  (where-is-internal 'newline-and-indent))
; PBE keysym syntax differs in xemcas
; (define-key lard-mode-map [RET] 'lard-newline-and-indent)
; PAR use same keysym syntax for both X and Gnu emacsen
;  (define-key lard-mode-map 'return 'lard-newline-and-indent)
;  (define-key lard-mode-map ")" 'lard-electric-paren))
  (define-key lard-mode-map (read-kbd-macro "RET") 'lard-newline-and-indent)
  (define-key lard-mode-map (read-kbd-macro ")") 'lard-electric-paren)
  )

(defvar lard-mode-syntax-table nil
  "Syntax table used in `lard-mode' buffers.")

(if lard-mode-syntax-table
    ()
  (setq lard-mode-syntax-table (make-syntax-table))
  (mapcar (function
	   (lambda (x) (modify-syntax-entry
			(car x) (cdr x) lard-mode-syntax-table)))
	  '(
	    ( ?\$ . "_" ) ( ?\% . "_" )
	    ( ?\/ . ". 14" ) (?\* . ". 23") 
	    ( ?\= . "_" )
	    ( ?\| . "_" ) (?\\ . "_" )
	    ( ?\_ . "_" )	
	    ( ?\' . "/")	
	    ( ?\" . "\"" )
	    ( ?\n . " ")
	    ))) ; ugh

(defconst lard-stringlit-re
   "\"\\([^\"\n\\]\\|\\\\.\\)*\""	; double-quoted
  "Regexp matching a Lard string literal.")



(defconst lard-blank-or-comment-re "[ \t]*\\($\\|/*\\)" ; ugh
  "Regexp matching blank or comment lines.")

(defconst lard-imenu-generic-re "[\_]([a-zA-Z_][a-zA-Z0-9_]*\\):var("
  "Regexp matching Identifier definitions.")

;; Sadly we need this for a macro in Emacs 19.
(eval-when-compile
  ;; Imenu isn't used in XEmacs, so just ignore load errors.
  (condition-case ()
      (require 'imenu)
    (error nil)))


(defvar lard-imenu-generic-expression
  (list (list nil lard-imenu-generic-re 1))
  "Expression for imenu")

;;;###autoload
(defun lard-mode ()
  "Major mode for editing Lard code.

\\(lard-mode-map)

Usually, tab indents for Lard code.  \")\" possibly reindents the
current line (to handle hanging right parentheses in the style of the
example code distributed with Lard).

The variable `lard-indent-column' controls the number of column to
indent each logical level of Lard code.

Turning on Lard mode calls the value of the variable `lard-mode-hook',
if non-nil."
  (interactive)
  (kill-all-local-variables)
  (make-local-variable 'font-lock-defaults)
  (make-local-variable 'paragraph-separate)
  (make-local-variable 'paragraph-start)
  (make-local-variable 'require-final-newline)
  (make-local-variable 'comment-start)
  (setq comment-start "/*")
  (make-local-variable 'comment-end)
  (setq comment-end "*/")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "[\\_]\\(/\*\\)[\\_]")
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'comment-column)
  (make-local-variable 'imenu-generic-expression)
  (setq imenu-generic-expression lard-imenu-generic-expression)
  (make-local-variable 'indent-line-function)
  (set-syntax-table lard-mode-syntax-table)
  (setq major-mode             'lard-mode
	mode-name              "Lard"
	local-abbrev-table     lard-mode-abbrev-table
	font-lock-defaults     '(lard-font-lock-keywords nil nil
				 ((?\/ . ". 14" ) (?* . ". 23")))
	paragraph-separate     "^[ \t]*$"
	paragraph-start        "^[ \t]*$"
	require-final-newline  t
	indent-line-function	'lard-indent-relative
	)
  (use-local-map lard-mode-map)

  ;; run the mode hook.
  (if lard-mode-hook
      (run-hooks 'lard-mode-hook)))


(provide 'lard-mode)
;; lard-mode.el ends here


; PBE xemacs doesn't have indent-line-to
; this is copied from FSF GNU Emacs indent.el
; PAR some XEmacsen now have this, only define it if undefined

(if (fboundp 'indent-line-to) ()
  (defun indent-line-to (column)
    "Indent current line to COLUMN.
This function removes or adds spaces and tabs at beginning of line
only if necessary.  It leaves point at end of indentation."
    (back-to-indentation)
    (let ((cur-col (current-column)))
      (cond ((< cur-col column)
             (if (> (- column (* (/ cur-col tab-width) tab-width)) tab-width)
                 (delete-region (point)
                                (progn (skip-chars-backward " ") (point))))
             (indent-to column))
            ((> cur-col column) ; too far right (after tab?)
             (delete-region (progn (move-to-column column t) (point))
                            (progn (back-to-indentation) (point))))))))
