/*
    The Language for Asynchronous Research and Development -- LARD
    Copyright (C) 1996-1999 AMULET Group, Department of Computer Science
        The University of Manchester, Oxford Road, Manchester, UK, M13 9PL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Id: format.c,v 1.20 2002/03/12 15:58:39 bardslea Exp $

    
	Printf-like format string handling (formatting/parsing)

	11 Jun 1999, AMULET Group
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../config.h"

#ifdef HAVE_LIBGMP
#include <gmp.h>
#endif

#include <limits.h>

#include <alloc.h>
#include <simple.h>

#include "format.h"
#include "lierror.h"
#include "stacks.h"
#include "stringops.h"

#define LARD_ALLOW_C_CONST_PREFICES 1
#define LARD_ALLOW_UNDERSCORES_IN_NUMBERS 1

#define MACH_INT_BITS (CHAR_BIT * sizeof(int))

/* formatBuffer: string buffer used while formating strings */
static char formatBuffer[1024];

static char *make_spaces_string (int count)
{
	char *ret = (char *) palloc (count + 1);
	char *retPtr = ret;

	ret[count] = '\0';
	while (count >= 0)
	{
		*retPtr = ' ';
		retPtr ++;
		count --;
	}
	return ret;
}

/* String helpers */
static char *end_of_string (char *str) { while (*str) str ++; return str; }
static char *skip_ws (char *str) { while (*str && isspace (*str)) str ++; return str; }
static char *remove_char_from_string (char *str, char chr)
{
	char *toPtr = str;
	char *fromPtr = str;

	while (*fromPtr)
	{
		if (*fromPtr != chr)
		{
			*toPtr = *fromPtr;
			toPtr ++;
		}
		fromPtr ++;
	}
	*toPtr = '\0';
	return str;
}
static void insert_char (char **str, char chr) { **str = chr; (*str)[1] = '\0'; (*str)++; }
/* digit_value: return the base 2..36 value of the given digit character or -1 for an invalid digit */
static int digit_value (char digit)
{ return (isdigit (digit) ? digit - '0' : (isalpha (digit) ? 10 + (tolower (digit)) - 'a' : -1)); }
/* parse_number: return the number under the pointer *str and advance *str */
static int parse_number (char **str) { int ret = atoi (*str); while (isdigit (**str)) (*str) ++; return ret; }
/* parse_lard_radix: parse a radix prefix from the given string.  If non is present then return
    the default radix `radix', don't check for octal prefix unless radix == 0 */
static int parse_lard_radix (char **str, int radix)
{
    char *numPtr = *str;
    bool firstDigitWasZero;
    if (radix > 36) { li_fatal ("parse_lard_radix: invalid radix (%d) must be <= 36", radix); return -10; }

    firstDigitWasZero = (*numPtr == '0');
    /* find the end of the number string */
    while (isdigit (*numPtr)) numPtr++;
    /* use the # prefix number as the radix */
    if (*numPtr == '#') { *numPtr = '\0'; radix = atoi (*str); *numPtr = '#'; (*str) = numPtr + 1; }
#if LARD_ALLOW_C_CONST_PREFICES
    else if (firstDigitWasZero)
    {
    	numPtr = (*str) + 1;
    	switch (*numPtr)
    	{
		case 'b':
			radix = -2; (*str) = numPtr + 1;
			break;
		case 'x':
			radix = -16; (*str) = numPtr + 1;
			break;
		default:
			if (radix == 0) radix = -8;
			/* Don't change str */
			break;
		}
#endif
	}

    return radix;
}
/* parse_lard_number: return the number under the pointer *str and advance *str using the radix `radix'
    will only read as many as maxWidth characters */
static int parse_lard_number (char **str, int radix, int maxWidth)
{
    int ret = 0;

#if LARD_ALLOW_UNDERSCORES_IN_NUMBERS
    while ((isalnum (**str) || **str == '_') && maxWidth) if (**str != '_')
#else
    while (isalnum (**str) && maxWidth)
#endif
    {
        int digitVal = digit_value (**str);
        if (digitVal >= radix)
        { 
            li_fatal ("parse_lard_number: digit `%c' is not valid in radix %d",
                **str, radix);
            return 0;
        }
        (*str) ++;
        maxWidth --;
        ret = (ret * radix) + digitVal;
    }
#if LARD_ALLOW_UNDERSCORES_IN_NUMBERS
	else { (*str) ++; maxWidth --; }
#endif
    return ret;
}
/* parse_ident: return the length of the ident under the pointer str */
static int parse_ident (char *str) { int ctr = 0; while (*str == '_' || isalnum (*str)) { ctr ++; str ++; } return ctr; }
/* radix_string: append a radix `radix' representation of `value' to string *str.  Returns count
    of written bytes (pads to at least `minWidth' bits with character `padChar' unless `minWidth' < 0 */
static char *radix_chars = "0123456789abcdef";
static int radix_string (char **str, unsigned value, int radix, int minWidth, char padChar)
{
	char *startPtr = *str;
	char *endPtr;
	int charTotal = 0;
	if (value == 0) { **str = '0'; (*str) ++; charTotal = 1; }

	while (value != 0)
	{
		**str = radix_chars[value % radix];
		value /= radix;
		(*str) ++;
		charTotal ++;
	}
	while (minWidth > charTotal) /* pad to minWidth if nec. */
	{
		**str = padChar;
		(*str) ++;
		charTotal ++;
	}
	**str = '\0';

	/* Now reverse the string */
	endPtr = *str - 1; /* last `real' char */
	while (startPtr < endPtr)
	{
		char tmp = *startPtr;
		*startPtr = *endPtr;
		*endPtr = tmp;

		startPtr ++;
		endPtr --;
	}
    return charTotal;
}

/* {mpint_,}bit_field: bit field extraction ops for int and mpz_t,
    sign extend bit max(left,right)-1 through the rest of the word if `isSigned' is true */
static int bit_field (int val, int left, int right, bool signExtend)
{
    int width = abs (left - right) + 1;
    int mask = (width == MACH_INT_BITS ? -1 : ((1 << width) - 1));
    int ret = (val >> min (left, right)) & mask;
    return (signExtend && ((1 << (width - 1)) & ret) ? (~mask | ret) : ret);
}
#ifdef HAVE_LIBGMP
static mpz_ptr mpint_bit_field (mpz_ptr val, int left, int right, bool signExtend)
{
    int width = abs (left - right) + 1;
    int offset = min (left, right);
    mpz_ptr ret = palloc (sizeof (mpz_t));
    mpz_t mask;
    mpz_init_set_ui (mask, 0);
    mpz_init_set_ui (ret, 0);
    mpz_setbit (mask, width);
    mpz_sub_ui (mask, mask, 1);
    mpz_div_2exp (ret, val, offset);
    mpz_and (ret, ret, mask);
    if (signExtend)
    {
        mpz_div_2exp (mask, mask, 1);
        if (mpz_cmp (ret, mask) > 0)
        {
            mpz_set_ui (mask, 0);
            mpz_setbit (mask, width);
            mpz_sub (ret, mask, ret);
            mpz_neg (ret, ret);
        }
    }
    mpz_clear (mask);
    return ret;
}
#endif

/* insert_bit_field: insert the value `insertVal' into the value `val' replacing
    the bit sequence [left:right].  The indices can be in either order
    and the inserted value is clipped.  Returns the modified `val'. */
static int insert_bit_field (int val, int insertVal, int left, int right)
{
    int width = abs (left - right) + 1;
    int offset = min (left, right);
    int mask = (width == MACH_INT_BITS ? -1 : ((1 << width) - 1));

    return (val & (~mask << offset)) | ((insertVal & mask) << offset);
}
#ifdef HAVE_LIBGMP
/* NB. modifies val */
static void mpint_insert_bit_field (mpz_ptr val, mpz_ptr insertVal, int left, int right)
{
    int width = abs (left - right) + 1;
    int offset = min (left, right);
    mpz_t tmp, mask;
    mpz_init (tmp);
    mpz_init_set_ui (mask, 0);
    mpz_setbit (mask, width);
    mpz_sub_ui (mask, mask, 1);
    /* Clip the parsed int */
    mpz_and (tmp, mask, insertVal);
    mpz_mul_2exp (tmp, tmp, offset);
    /* Shift the mask, cut a hole in the argument mpint */
    mpz_mul_2exp (mask, mask, offset);
    mpz_com (mask, mask);
    mpz_and (val, val, mask);
    /* Insert the clipped value */
    mpz_ior (val, val, tmp);
    mpz_clear (tmp);
    mpz_clear (mask);
}
#endif

/* lard_format_from_array: make a formatted string from the args in `args' */
char *lard_format_from_array (char *format, int argCount, void **args)
{
    char *ret = formatBuffer;
    int argIndex = 0;
    bool looping = FALSE;
    char *repeatedFormat = NULL; /* the repeatable format string */
    char *separator = NULL; /* repeated format separator */

    *ret = '\0';

    while (*format || looping)
    {
        int minWidth = -1; /* No width limit */
        bool pad = FALSE; /* Don't zero pad */
        bool advanceArgs = TRUE; /* Move from arg to arg */
        bool isMPInt = FALSE; /* Arg is an mpint */
        int leftIndex = -1; /* Don't use indices */
        int rightIndex = -1;
        bool isSigned = FALSE; /* - prefix, treat %d options as signed */

        if (looping) /* search for start of format */
            { format = repeatedFormat; }
        else {
            while (*format && *format != '%') { insert_char (&ret, *format); format ++; }
            if (! *format) return strdup (formatBuffer);
            if (! format[1]) { li_fatal ("formatdata: invalid format string\n"); return NULL; }
            format ++;
        }

        if (*format == '*') /* repeat to end of format */
        {
            format++;
            if (*format != '(') /* char separator */
            {
                separator = palloc (2);
                separator[0] = *format; separator[1] = '\0';
                format++;
            } else {
                int sepLength = parse_ident (format + 1);
                if (format[sepLength + 1] != ')')
                    { li_fatal ("formatdata: expecting `)' in format\n"); return NULL; }
                *format = '\0';
                separator = palloc (sepLength + 1);
                strncpy (separator, format + 1, sepLength);
                format += sepLength + 2;
            }
            looping = TRUE;
            repeatedFormat = format;
        }
        /* don't skip to next format */
        if (*format == '!') { advanceArgs = looping; format ++; } /* Don't comsume arg */
        if (*format == '-') { isSigned = TRUE; format ++; } /* do a signed extract for %d */
        if (isdigit (*format)) /* [0][int] width/pad prefix */
        {
            if (*format == '0') { pad = TRUE; format ++; }
            minWidth = parse_number (&format);
        }
        if (*format == '(') /* bit fields/word selector */
        {
            int wordNo;
            format ++; /* skip ( */
            wordNo = parse_number (&format);

            switch (*format)
            {
                case ',': /* was the word index */
                    argIndex = wordNo;
                    if (argIndex >= argCount)
                        { li_fatal ("formatdata: argument index `%d' too big", argIndex); return NULL; }
                    format ++;
                    leftIndex = parse_number (&format);
                    if (*format != ':') { li_fatal ("formatdata: expecting `:'\n"); return NULL; }
                    format ++;
                    rightIndex = parse_number (&format);
                    break;
                case ':': /* was the left index */
                    format ++;
                    leftIndex = wordNo;
                    rightIndex = parse_number (&format);
                    break;
                case ')': /* end of address, don't skip */
                    argIndex = wordNo;
                    break;
                default:
                    li_fatal ("formatdata: unexpected char `%c'\n", *format);
                    return NULL;
                    break;
            }
            if (*format != ')') { li_fatal ("formatdata: expecting `)' in format\n"); return NULL; }
            format ++;
        }
        if (*format == 'm')
        {
#ifdef HAVE_LIBGMP
            isMPInt = TRUE;
            format ++;
#else
            li_fatal ("formatdata: this installation of LARD was compiled without mpint support\n");
            return NULL;
#endif
        }
        switch (*format)
        {
            case 'd': case 'o': case 'x': case 'X': case 'b': case 'u':
            case 'i':
				{
					int radix = (*format == 'd' ? -10
						: (*format == 'i' ? -10
						: (*format == 'u' ? 10
						: (*format == 'o' ? 8
						: (*format == 'b' ? 2
						: 16 )))));
#ifdef HAVE_LIBGMP
					if (isMPInt)
					{
						mpz_ptr val = (leftIndex >= 0
							? mpint_bit_field ((mpz_ptr) args[argIndex], leftIndex, rightIndex, isSigned)
							: (mpz_ptr) args[argIndex]);
						int argWidth = mpz_sizeinbase (val, abs (radix));
						if (minWidth > 0 && argWidth < minWidth)
						{
							minWidth -= argWidth;
							while (minWidth) { insert_char (&ret, (pad ? '0' : ' ')); minWidth--; }
						}
						mpz_get_str (ret, abs (radix), val);
						/* free the bit field mpint */
						if (leftIndex >= 0) { mpz_clear (val); free (val); }
					} else
#endif
					{ /* simple integer */
						int val = (leftIndex >= 0
							? bit_field ((int) args[argIndex], leftIndex, rightIndex, isSigned)
							: (int) args[argIndex]);
						char *retPtr = ret;

						if (radix < 0 && val < 0) /* print the leading - */
						{
							*retPtr = '-'; retPtr++;
							val = -val;
						}
						radix_string (&retPtr, val, abs (radix), minWidth, (pad ? '0' : ' '));
					}
					while (*ret) { if (*format == 'X') *ret = toupper (*ret); ret++; }
					format ++;
				}
				break;
            case 's': /* strings */
                {
                    int argWidth;
                    struct crec *cr = (struct crec *) args[argIndex];
                    if (checkstrcr (cr, "format"))
                    {
                        if (leftIndex >= 0)
                            { li_fatal ("formatdata: can't print a bitfield as a string\n"); return NULL; }
                        argWidth = strlen (cr->str);
                        if (minWidth > 0 && argWidth < minWidth)
                            { minWidth -= argWidth; while (minWidth) { insert_char (&ret, ' '); minWidth --; } }
                        strcpy (ret, cr->str);
                    }
                    format ++;
                } break;
            case 'c': /* char */
                format ++;
#ifdef HAVE_LIBGMP
                if (isMPInt)
                {
                    mpz_ptr val = (leftIndex >= 0
                        ? mpint_bit_field ((mpz_ptr) args[argIndex], leftIndex, rightIndex, isSigned)
                        : (mpz_ptr) args[argIndex]);
                    insert_char (&ret, (char) mpz_get_ui (val));
                    if (leftIndex >= 0) { mpz_clear (val); free (val); }
                } else
#endif
                {
                    int val = (leftIndex >= 0
                        ? bit_field ((int) args[argIndex], leftIndex, rightIndex, isSigned)
                        : (int) args[argIndex]);
                    insert_char (&ret, (char) val);
                } break;
            case 'e': /* enumeration */
                {
#ifdef HAVE_LIBGMP
                    mpz_ptr mpVal = NULL;
#endif
                    int intVal = 0, identLength, elementVal = 0;
                    bool matched = FALSE;
#ifdef HAVE_LIBGMP
                    if (isMPInt) /* So field extraction */
                    {
                        mpVal = (leftIndex >= 0
                            ? mpint_bit_field ((mpz_ptr) args[argIndex], leftIndex, rightIndex, isSigned)
                            : (mpz_ptr) args[argIndex]);
                    } else
#endif
                    {
                        intVal = (leftIndex >= 0
                            ? bit_field ((int) args[argIndex], leftIndex, rightIndex, isSigned)
                            : (int) args[argIndex]);
                    }
                    format ++;
                    if (*format != '(') { li_fatal ("formatdata: expecting `('\n"); return NULL; }
                    format ++;

                    do {
                        char *ident = format;
                        identLength = parse_ident (format);
                        if (identLength == 0) { li_fatal ("formatdata: expecting an identifier\n"); return NULL; }
                        format += identLength;
                        if (*format == '=')
						{
							format ++;
							if (*format == 'e') { matched = TRUE; format++; }
							else elementVal = parse_number (&format);
						}
                        if (matched || (isMPInt
#ifdef HAVE_LIBGMP
                            ? mpz_cmp_ui (mpVal, elementVal) == 0
#else
                            FALSE
#endif
                            : intVal == elementVal))
                        { /* Found a match */
                            char sep = ident[identLength];
                            ident[identLength] = '\0';
                            if (minWidth > 0 && identLength < minWidth)
                            {
                                minWidth -= identLength;
                                while (minWidth) { insert_char (&ret, ' '); minWidth --; }
                            }
                            strcpy (ret, ident);
                            ident[identLength] = sep;
                            /* Skip to end */
                            while (*format && *format != ')') format ++;
                            matched = TRUE;
                        }
                        if (*format == ',') format ++;
                        else if (*format == ')') { /* End */ }
                        else { li_fatal ("formatdata: %c expecting `,' or `)'\n", *format); return NULL; }
                        elementVal ++;
                    } while (*format != ')' && ! matched);
                    if (! matched) /* No match, show value in brackets */
                    {
#ifdef HAVE_LIBGMP
                        if (isMPInt) { insert_char (&ret, '('); mpz_get_str (ret, 10, mpVal); strcat (ret, ")"); }
						else
#endif
                        {
							insert_char (&ret, '(');	
							radix_string (&ret, intVal, 10, 0, ' ');
							strcat (ret, ")");
						}
                    }
                    format ++;
#ifdef HAVE_LIBGMP
                    if (isMPInt && leftIndex >= 0) { mpz_clear (mpVal); free (mpVal); }
#endif
                } break;
            case '%': /* emit % */
                format ++;
                insert_char (&ret, '%');
                advanceArgs = FALSE;
                break;
            case '\0':
                li_fatal ("formatdata: invalid format string"); return NULL;
                break;
        }
        if (advanceArgs) argIndex ++;
        if (looping && argIndex < argCount) /* add separator to repeated format */
            { strcat (ret, separator); }
        ret = end_of_string (ret);
        if (argIndex >= argCount) break;
    }
    /* print rest of format string verbatim if we looped on one item */
    if (*format)
        { strcpy (ret, format); }

    return strdup (formatBuffer);
}

/* make_format_array: make an array of void* from a string containing
    hex values and maybe some space separation */
#define HEX_DIGITS_PER_WORD (MACH_INT_BITS / 4)
void **make_format_array (int argCount, char *argStr)
{
    void **argArray = palloc (sizeof (void *) * argCount);
    char tmpChar;
    int argIndex = 0;
    argStr = skip_ws (argStr);

    while (argIndex < argCount && *argStr)
    {
        tmpChar = argStr[HEX_DIGITS_PER_WORD];
        argStr[HEX_DIGITS_PER_WORD] = '\0';
        sscanf (argStr, "%X", (int *) (argArray + argIndex));
        argStr[HEX_DIGITS_PER_WORD] = tmpChar;
        argStr += HEX_DIGITS_PER_WORD;
        argIndex ++;
    }
    if (argCount != argIndex)
        li_fatal ("make_format_array: %d too few args\n", (argCount - argIndex));
    return argArray;
}

/* formatdata: format some raw data according to a LARD format string */
void formatdata (void)
{
    struct crec *format = (struct crec *) param_pop ();
    struct crec *ret;
    char *formatString;
    int wordCount = (int) param_pop () / sizeof (int);
    void **data = palloc (sizeof (void *) * wordCount);
    int i;

    if (! checkstrcr (format, "formatdata")) return;
    formatString = format->str;
    ret = palloc (sizeof (struct crec));
    ret->refcount = 0;

    for (i = 0; i < wordCount; i++) data[i] = (void *) param_pop ();

    ret->str = lard_format_from_array (formatString, wordCount, data);

    free_if_zero (format);

    param_push ((int) ret);
    free (data);
}

/* lard_parse_to_array : parse the given string using the given format
    and put the parsed values in addresses/words pointed to by args */
char *lard_parse_to_array (char *format, char *str, int argCount, void **args)
{
    char *ret = str;
    int argIndex = 0;
    bool looping = FALSE;
    char *repeatedFormat = NULL; /* the repeatable format string */
    char *separator = NULL; /* repeated format separator */
    int defaultRadix = 0; /* default input radix for %i, 0 means `decimal but accept any base prefix' */
    	/* NB. octal 0 prefix is only accepted with radix of 0 */
    char tempChar;
    char *tempCharPos;
	char *startOfFormat = format;

    while (*format || looping)
    {
        int maxWidth = -1; /* No width limit */
        bool pad = FALSE; /* Don't zero pad */
        bool advanceArgs = TRUE; /* Move from arg to arg */
        bool overwriteBitfield = FALSE; /* TRUE => overwrite target on bitfield insert, FALSE => insert */
        bool isMPInt = FALSE; /* Arg is an mpint */
        int leftIndex = -1; /* Don't use indices */
        int rightIndex = -1;
        bool isSigned = FALSE; /* - prefix, treat %d options as signed */

        ret = skip_ws (ret);
        if (looping) /* search for start of format */
            { format = repeatedFormat; }
        else {
            /* skip whitespace then do a literal match */
            format = skip_ws (format);
            while (*ret && *format && *format != '%')
            {
                if (*format != *ret)
                {
                    li_fatal ("parsedata: invalid character, expecting `%c' got `%c':\n`%s'\n%s^\n",
						*format, *ret, str, make_spaces_string ((int) (ret - str)));
                    return NULL;
                }
                format ++;
                format = skip_ws (format);
                ret ++;
                ret = skip_ws (ret);
            }
            if (! *format || ! *ret) return strdup (ret);
            if (! format[1])
			{
				li_fatal ("parsedata: invalid format string:\n`%s'\n%s^\n",
					startOfFormat, make_spaces_string ((int) (format - startOfFormat)));
				return NULL;
			}
            format ++;
        }

        if (*format == '*') /* repeat to end of format */
        {
            format++;
            if (*format != '(') /* char separator */
            {
                separator = palloc (2);
                separator[0] = *format; separator[1] = '\0';
                format++;
            } else {
                int sepLength = parse_ident (format + 1);
                if (format[sepLength + 1] != ')')
                    { li_fatal ("parsedata: expecting `)' in format\n"); return NULL; }
                *format = '\0';
                separator = palloc (sepLength + 1);
                strncpy (separator, format + 1, sepLength);
                format += sepLength + 2;
            }
            looping = TRUE;
            repeatedFormat = format;
        }
        /* don't skip to next format */
        if (*format == '!') { advanceArgs = looping; format ++; } /* Don't comsume arg */
        if (*format == '=') { overwriteBitfield = TRUE; format ++; } /* Overwrite bitfield don't insert */
        if (*format == '-') { isSigned = TRUE; format ++; } /* do a signed extract for %d */
        if (isdigit (*format)) /* [0][int] width/pad prefix */
        {
            if (*format == '0') { pad = TRUE; format ++; }
            maxWidth = parse_number (&format);
        }
        if (*format == '(') /* bit fields/word selector */
        {
            int wordNo;
            format ++; /* skip ( */
            wordNo = parse_number (&format);

            switch (*format)
            {
                case ',': /* was the word index */
                    argIndex = wordNo;
                    if (argIndex >= argCount)
                        { li_fatal ("parsedata: argument index `%d' too big", argIndex); return NULL; }
                    format ++;
                    leftIndex = parse_number (&format);
                    if (*format != ':') { li_fatal ("parsedata: expecting `:'\n"); return NULL; }
                    format ++;
                    rightIndex = parse_number (&format);
                    break;
                case ':': /* was the left index */
                    format ++;
                    leftIndex = wordNo;
                    rightIndex = parse_number (&format);
                    break;
                case ')': /* end of address, don't skip */
                    argIndex = wordNo;
                    break;
                default:
                    li_fatal ("parsedata: unexpected char `%c'\n", *format);
                    return NULL;
                    break;
            }
            if (*format != ')') { li_fatal ("parsedata: expecting `)' in format\n"); return NULL; }
            format ++;
        }
        if (*format == 'm')
        {
#ifdef HAVE_LIBGMP
            isMPInt = TRUE; format ++;
#else
            li_fatal ("parsedata: this installation of LARD was compiled without mpint support\n");
            return NULL;
#endif
        }
        switch (*format)
        {
            case 'd': case 'o': case 'x': case 'X': case 'b': case 'u':
            case 'i': {
                int radix = (*format == 'i' ? defaultRadix 
                    : (*format == 'd' ? -10
                    : (*format == 'u' ? 10
                    : (*format == 'o' ? -8
                    : (*format == 'b' ? -2
                    : -16 )))));
                int absRadix;
                int val;
                bool negate = FALSE;

                if (maxWidth > 0 && strlen (ret) > maxWidth) tempCharPos = ret + maxWidth;
                else {
                    tempCharPos = ret;
                    while (isalnum (*tempCharPos) || *tempCharPos == '#' || *tempCharPos == '-'
#if LARD_ALLOW_UNDERSCORES_IN_NUMBERS
						|| *tempCharPos == '_'
#endif
					) tempCharPos ++;
                }
                tempChar = *tempCharPos;
                *tempCharPos = '\0'; /* Insert a temporary null to prevent over-reading */

                if (radix <= 0 && *ret == '-')
                    { negate = TRUE; ret ++; } /* handle - prefix */
                radix = parse_lard_radix (&ret, radix);
                if (radix > 0 && negate)
    				{ li_fatal ("parsedata: - prefix found on number expected to be unsigned"); return NULL; }
                /* correct radix here */
                if (radix == 0) radix = -10;
                absRadix = abs (radix);
#ifdef HAVE_LIBGMP
                if (isMPInt)
                {
                    mpz_t tmp;
#if LARD_ALLOW_UNDERSCORES_IN_NUMBERS
                    char *deUnderScoredString = remove_char_from_string (strdup (ret), '_');
                    mpz_init_set_str (tmp, deUnderScoredString, absRadix);
                    free (deUnderScoredString);
#else
                    mpz_init_set_str (tmp, ret, absRadix);
#endif
                    if (negate) mpz_neg (tmp, tmp);

                    if (leftIndex >= 0) /* Need to do a bitfield insert */
                    {
                        if (overwriteBitfield) mpz_set_ui ((mpz_ptr) args[argIndex], 0);
                        mpint_insert_bit_field ((mpz_ptr) args[argIndex], tmp, leftIndex, rightIndex);
                    } else mpz_set ((mpz_ptr) args[argIndex], tmp);

                    /* Clean up */
                    mpz_clear (tmp);
                } else
#endif
                { /* simple integer */
                    val = parse_lard_number (&ret, absRadix, -1);

                    if (negate) val = -val;

                    if (leftIndex >= 0) /* Insert the val */
                    {
                        if (overwriteBitfield) args[argIndex] = (void *) 0;
                        args[argIndex] = (void *) insert_bit_field ((int) args[argIndex], val, leftIndex, rightIndex);
                    } else args[argIndex] = (void *) val;
                }
                *tempCharPos = tempChar;
                do {
                    int digitVal = digit_value (*ret);
#if LARD_ALLOW_UNDERSCORES_IN_NUMBERS
                    if (*ret == '_' || (digitVal != -1 && digitVal < absRadix)) ret ++;
#else
                    if (digitVal != -1 && digitVal < absRadix) ret ++;
#endif
                    else break;
                } while (*ret);
                format ++;
            }
            break;
        case 's': /* strings */
            {
                struct crec *cr = (struct crec *) args[argIndex];
                if (cr)
                {
                    if (leftIndex >= 0)
                        { li_fatal ("parsedata: can't print a bitfield as a string\n"); return NULL; }
                    /* deallocate old string */
                    free_if_zero (cr);
                    /* count out the characters to go into the string */
                    tempCharPos = ret;
                    while (*tempCharPos && (maxWidth > 0 ? tempCharPos <= (ret + maxWidth) : TRUE)
                        && ! isspace (*tempCharPos)) tempCharPos++;
                    tempChar = *tempCharPos;
                    *tempCharPos = '\0'; /* Insert a temporary null to prevent reading non digit chars */
                    ((struct crec *) args[argIndex])->refcount --;
                    free_if_zero ((struct crec *) args[argIndex]);
					{
						struct crec *newString = palloc (sizeof (struct crec));

						newString->str = strdup (ret);
						newString->refcount = 1;
						args[argIndex] = newString;
					}
                    /* ((struct crec *) args[argIndex])->str = strdup (ret); */
                    ret = tempCharPos;
                    *tempCharPos = tempChar;
                } else {
					li_fatal("parsedata: uninitialised string (newstringvar not called?)\n");
					return NULL;
				}
                format ++;
            } break;
        case 'c': /* char */
            format ++;
#ifdef HAVE_LIBGMP
            if (isMPInt)
            {
                mpz_t tmp;
                mpz_init_set_ui (tmp, (unsigned char) *ret);

                if (leftIndex >= 0) /* Need to do a bitfield insert */
                {
                    if (overwriteBitfield) mpz_set_ui ((mpz_ptr) args[argIndex], 0);
                    mpint_insert_bit_field ((mpz_ptr) args[argIndex], tmp, leftIndex, rightIndex);
                } else mpz_set ((mpz_ptr) args[argIndex], tmp);

                /* Clean up */
                mpz_clear (tmp);
            } else
#endif
            {
                int val = (unsigned char) *ret;
                if (leftIndex >= 0) /* Insert the val */
                {
                    if (overwriteBitfield) args[argIndex] = (void *) 0;
                    args[argIndex] = (void *) insert_bit_field ((int) args[argIndex], val, leftIndex, rightIndex);
                } else args[argIndex] = (void *) val;
            }
            ret ++; break;
        case 'e': /* enumeration */
            {
                int identLength, elementVal = 0, matchedValue = 0, matchedIdentLength = 0;
                bool matched = FALSE;
                format ++;
                if (*format != '(') { li_fatal ("parsedata: expecting `('\n"); return NULL; }
                format ++;

                if (isdigit (*ret))
                {
                    matchedValue = parse_lard_number (&ret, defaultRadix, maxWidth);
                    matched = TRUE;
                } else do {
                    char *ident = format;
                    identLength = parse_ident (format);
                    if (identLength == 0) { li_fatal ("parsedata: expecting an identifier\n"); return NULL; }
                    format += identLength;
                    if (*format == '=')
					{
						format ++;
						if (*format == 'e') { elementVal = 0; format ++; }
						else elementVal = parse_number (&format);
					}

                    if (strncmp (ident, ret, identLength) == 0)
                    { /* Found a match */
                        if (! matched || identLength > matchedIdentLength)
                        {
                            matchedIdentLength = identLength;
                            matchedValue = elementVal;
                            matched = TRUE;
                        }
                    }
                    if (*format == ',') format ++;
                    else if (*format == ')') { /* End */ }
                    else { li_fatal ("parsedata: expecting `,' or `)' in %%e format\n"); return NULL; }
                    elementVal ++;
                } while (*format != ')'/* && ! matched */);

                /* skip to end of format */
                while (*format != ')') format ++;

                if (matched)
                {
#ifdef HAVE_LIBGMP
                    if (isMPInt)
                    {
                        mpz_t tmp;
                        mpz_init_set_si (tmp, matchedValue);

                        if (leftIndex >= 0) /* Need to do a bitfield insert */
                        {
                            if (overwriteBitfield) mpz_set_ui ((mpz_ptr) args[argIndex], 0);
                            mpint_insert_bit_field ((mpz_ptr) args[argIndex], tmp, leftIndex, rightIndex);
                        } else mpz_set ((mpz_ptr) args[argIndex], tmp);

                        /* Clean up */
                        mpz_clear (tmp);
                    } else
#endif
                    {
                        if (leftIndex >= 0) /* Insert the val */
                        {
                            if (overwriteBitfield) args[argIndex] = (void *) 0;
                            args[argIndex] = (void *) insert_bit_field ((int) args[argIndex],
                                matchedValue, leftIndex, rightIndex);
                        } else args[argIndex] = (void *) matchedValue;
                    }
                    ret += matchedIdentLength;
                } else /* No match */
                    { li_fatal ("parsedata: invalid enumeration value"); return NULL; }
                format ++;
            } break;
        case '%': /* emit % */
            if (*format != *ret)
                li_fatal ("parsedata: invalid character, expecting `%c' got `%c'\n", *format, *ret);
            format ++;
            ret++;
            advanceArgs = FALSE;
            break;
        case 'r': /* set radix */
            format ++;
            defaultRadix = (int) args[argIndex];
            break;
        case '\0':
            li_fatal ("parsedata: invalid format string"); return NULL;
            break;
        }
        if (advanceArgs) argIndex ++;
        if (looping && argIndex < argCount)
        { /* check separator */
            char *sepPtr = separator;
            ret = skip_ws (ret);
            while (*ret && *sepPtr)
            {
                if (*ret != *sepPtr) { li_fatal ("parsedata: expecting `%s' separator", separator); return NULL; }
                ret ++; sepPtr ++;
            }
            if (! *ret && *separator) { li_fatal ("parsedata: end of string while parsing separator"); return NULL; }
        }
        if (argIndex >= argCount) break;
    }
    /* OK, check the rest of the format verbatim */
    format = skip_ws (format);
    ret = skip_ws (ret);
    while (*format && *ret)
    {
        if (*ret && *format != *ret)
        {
            li_fatal ("parsedata: expecting character `%c' got `%c'", *format, *ret); return NULL;
        } else {
            format ++; ret ++;
        }
        format = skip_ws (format);
        ret = skip_ws (ret);
    }
    if (*format) { li_fatal ("parsedata: expecting `%s' at end of format", format); return NULL; }

    return strdup (ret);
}

/* parsedata: parse some raw data according to a LARD format string,
    this function will be called with a single ref/var pointing to
    the data and a data byte count */
void parsedata (void)
{
    struct crec *format = (struct crec *) param_pop ();
    struct crec *str = (struct crec *) param_pop ();
    struct crec *ret;
    char *formatString;
    char *stringToParse;
    int wordCount = (int) param_pop () / sizeof (int);
    void *data = (void **) param_pop ();

    if (! checkstrcr (format, "parsedata")) return;
    formatString = format->str;
    stringToParse = str->str;
    ret = palloc (sizeof (struct crec));
    ret->refcount = 0;

    ret->str = lard_parse_to_array (formatString, stringToParse, wordCount, data);
    if (! ret->str) ret->str = strdup ("");

    free_if_zero (format);
    free_if_zero (str);

    param_push ((int) ret);
}

