/*
    The Language for Asynchronous Research and Development -- LARD
    Copyright (C) 1996-1999 AMULET Group, Department of Computer Science
        The University of Manchester, Oxford Road, Manchester, UK, M13 9PL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Id: sparray.c,v 1.2 2000/04/25 00:16:21 bardslea Exp $

    
	Sparse array functions for LARD

	11 Jun 1999, AMULET Group
*/

#include <stdlib.h>
#include <stdio.h>

#include <alloc.h>
#include <simple.h>

#include "lierror.h"
#include "sparray.h"
#include "stacks.h"

/* Sparse array functions for LARD */

/* A sparse array is the sort of structure that we need for memory
   models and the like.  It's like Perl's associative arrays, except
   that the index is always an integer. */

/* Apparently all the answers are in Knuth: The Art of Computer
   Programming, part 3.  Not suprising. */

/* We don't want to impose an initialisation-time limit on the maximum
   number of items in the array.  This inevitably means that we have
   to periodically "rehash" the array by creating a new larger array
   and copying all the data across. */

/* There are some "magic" variables and functions that can be fiddled
   with to get this to work efficiently: */

/* maxfullness(n) is the number of elements of an n-element array that
   we allow to become used before rehashing.  We need to be sure that
   the array never becomes 100% full.  */

#define maxfullness(n) (n-n/20-1)

/* hash_constant is a large number relatively prime to 2^32 used in
   the hash function.  See comments there. */

#define hash_constant (0x3fffffff)

/* grow_size(n) is the size of the new larger table when a table of
   size n overflows. */

#define grow_size(n) (n+n/10+1)



/* debug messages */

#define spdbg(format, args...) ;
/* #define spdbg printf */


/* Sparse array is represented by a struct sparray, with global info
   and a pointer to the table.  The table is an array of struct
   spentrys, each of which has a pointer to its data.  We can't keep
   the data in the spentry because it gets moved around when we
   rehash, and LARD wouldn't like variables to move around under its
   feet, so we allocate some data space pointed to by the struct
   sparray.  We allocate from this space by incrementing the pointer,
   though a copy of the original is kept so we can free it. When we
   grow the array we keep a linked list of pointers to the old data
   areas so that they can be freed too. */

struct sparray {
  unsigned int sz;
  unsigned int free;
  struct spentry * entries;
  void * data_space;
  void * init_data_space;
  struct oldbits * oldbits;
};

struct oldbits {
  struct oldbits * oldbits;
  void * init_data_space;
};

struct spentry {
  bool full;
  unsigned int index;
  void * data;
};


unsigned int hash ( unsigned int n )
{
      /* The scattering function. */
      /* Take an integer n and return another integer, such that the
         resulting integers have a random distribution whatever the
         distribution of the inputs integers are. */
      /* This is clearly impossible in general, but we can do it if we
         assume that the input integers are random except for (a) they
         are limited to a sub-range of the full integer space, and (b)
         they may have some properties modulo something, e.g. they are
         all multiples of 4. */
      /* Based on Knuth, we multiply by a large constant which is
         relatively prime to 2^32, and get the result modulo 2^32. */
  
  return (n*hash_constant);
}


bool is_prime ( unsigned int n )
{
      /* return a boolean indicating if n is prime */
  
  unsigned int i;

  if (n%2==0) {
    return FALSE;
  }
  
  i=3;
  while (i*i < n && n%i != 0)
        i+=2;
  return n%i != 0;
}


unsigned int round_up_prime ( unsigned int n )
{
      /* return a prime >= n */

  if (is_prime(n))
  {
    return n;
  } else {
    return round_up_prime(n+1);
  }
}


struct sparray * create_sparray ( unsigned int sz, int elemsz )
{
      /* Create a sparse array with initial size >= sz to store
         elements of size elemsz. */
      /* We round sz up to the next prime. See find_sparray for why! */
      /* Since hashing performance drops off quickly as the array
         fills, we don't make all of the entries useable. */
  
  unsigned int primesz;
  unsigned int useablesz;
  struct sparray * sp;
  
  primesz=round_up_prime(sz);
  useablesz=maxfullness(primesz);
  spdbg("create_sparray: requested size=%d, allocating %d, useable %d.\n",sz,primesz,useablesz);
  sp=palloc(sizeof(struct sparray));
  sp->sz=primesz;
  sp->free=useablesz;
  sp->entries=calloc(primesz,sizeof(struct spentry));
  sp->data_space=calloc(useablesz,elemsz);
  sp->init_data_space=sp->data_space;
  if (!sp->entries || !sp->data_space)
  {
    li_fatal("calloc failed in create_sparray");
  }
  sp->oldbits=NULL;
  return sp;
}


void free_oldbits ( struct oldbits * oldbits )
{
  if (oldbits)
  {
    free_oldbits(oldbits->oldbits);
    free(oldbits->init_data_space);
  }
}

void destroy_sparray ( struct sparray * sp )
{
      /* free all the space used by the sparray. */
  
  spdbg("destroy_sparray\n");
  free_oldbits(sp->oldbits);
  free(sp->entries);
  free(sp->init_data_space);
  free(sp);
}


struct spentry * find_sparray ( struct sparray * sp, unsigned int index,
                                int elemsz )
{
      /* Find an entry in sparray sp for index index.  If index is
         already present in the array its entry is returned.  If it
         isn't present, an empty entry which it may use is returned.
         If the array is full it is grown. */
  
  unsigned int hashnum;
  unsigned int i,j;

  hashnum=hash(index);
      /* This is the scatter-function that should remove any bias in the
         distribution of indicies. */
  i=hashnum%sp->sz;
      /* i is the initial array index for index. */

  if (!sp->entries[i].full || sp->entries[i].index==index) {
    return &(sp->entries[i]);
  }

  j=1+i%(sp->sz-2);
      /* If we don't find index at element i, we look at i+j, i+2j,
         i+3j etc. all mod size. Since sz is prime, for any j (except
         0 and sz) we know that we will eventually look in ALL
         elements. */

  while (sp->entries[i].full && sp->entries[i].index!=index) {
    i=(i+j)%sp->sz;
  }
  return &(sp->entries[i]);
}


void grow_sparray ( struct sparray * sp, int elemsz )
{
      /* sparray is full.  Increase its size, allocating a new array
         of spentrys, and transfer the data.  The actual "data" isn't
         copied; the pointers in the new spentries point to the old
         data.  We have to do this so that LARD doesn't see variables
         moving under its feet. */
  
  struct sparray oldsp;
  unsigned int newdata;
  int i;
  struct spentry * newentry;

  oldsp=*sp;

  sp->oldbits=palloc(sizeof(struct oldbits));
  sp->oldbits->oldbits=oldsp.oldbits;
  sp->oldbits->init_data_space=oldsp.init_data_space;
  sp->sz=round_up_prime(grow_size(oldsp.sz));
  newdata=maxfullness(sp->sz)-maxfullness(oldsp.sz);
  sp->data_space=calloc(newdata,elemsz);
  sp->init_data_space=sp->data_space;
  sp->free=newdata;
  sp->entries=calloc(sp->sz,sizeof(struct spentry));
  if (!sp->entries || !sp->data_space)
  {
    li_fatal("calloc failed in grow_sparray");
  }

  for(i=0;i<oldsp.sz;i++)
  {
    if (oldsp.entries[i].full)
    {
      newentry=find_sparray(sp,oldsp.entries[i].index,elemsz);
      newentry->full=TRUE;
      newentry->index=oldsp.entries[i].index;
      newentry->data=oldsp.entries[i].data;
    }
  }

  free(oldsp.entries);
}


void * index_sparray ( struct sparray * sp, unsigned int index, int elemsz )
{
      /* Look up index in sparray.  If it exists, return a pointer to
         its data.  If it doesn't, create a new entry and return a
         pointer to the new data. */

  struct spentry * entry;

      /* If sp is full, grow it.  We have to do this even though the
         index may already exist.  */

  if (sp->free==0)
  {
    grow_sparray(sp,elemsz);
  }
  
  entry=find_sparray(sp,index,elemsz);
  if (!entry->full)
  {
    entry->full=TRUE;
    entry->index=index;
    entry->data=sp->data_space;
    sp->data_space+=elemsz;
    sp->free--;
  }
  return entry->data;
}


void l_create_sparray ( void )
{
  struct sparray * sp;
  int T;
  struct sparray ** V;

  T=param_pop();
  V=(struct sparray **)param_pop();
  sp=create_sparray(17,T);
  (*V)=sp;
}


void l_destroy_sparray ( void )
{
  struct sparray * sp;
  int T;
  struct sparray ** V;

  T=param_pop();
  V=(struct sparray **)param_pop();
  sp=(*V);
  if (!sp)
  {
    li_fatal("Uninitialised sparse array passed to destroy_sparray");
  }
  destroy_sparray(sp);
}


void l_index_sparray ( void )
{
  int T;
  struct sparray ** V;
  unsigned int i;
  struct sparray * sp;
  void * data;
  
  T=param_pop();
  V=(struct sparray **)param_pop();
  i=(unsigned int)param_pop();
  sp=(*V);
  if (!sp)
  {
    li_fatal("Uninitialised sparse array passed to destroy_sparray");
  }
  data=index_sparray(sp,i,T);
  param_push((int)data);
}
